from scipy.stats import norm
import math


class sample_size:
    """
    Sample size calculation

    Parameters
    ----------
    type: Choose "mean" or "proportion"
    err_mean Acceptable Error for Mean
    sd_mean Standard deviation for Mean
    err_prop Acceptable Error for Proportion
    p_prop Initial proportion estimate for Proportion
    conf Confidence level
    pop_size Population size
    incidence Incidence rate (i.e., fraction of valid respondents)
    response Response rate

    @return A list of variables defined in sample_size as an object of class sample_size

    sample_size(type = "mean", err_mean = 2, sd_mean = 10)
    """

    def sample_size_mean(
        type,
        err_mean=2,
        sd_mean=10,
        err_prop=0.1,
        p_prop=0.5,
        conf=0.95,
        pop_size=None,
        incidence=1,
        response=1,
    ):
        if conf is None or conf < 0 or conf > 1:
            conf = 0.95

        zval = -norm.ppf((1 - conf) / 2)

        if type == "mean":
            if err_mean is None:
                return "Please select an acceptable error greater than 0"
            n = (zval**2 * sd_mean**2) / (err_mean**2)
        else:
            if err_prop is None:
                return "Please select an acceptable error greater than 0"
            n = (zval**2 * p_prop * (1 - p_prop)) / (err_prop**2)

        if pop_size is not None:
            n = n * pop_size / ((n - 1) + pop_size)

        return math.ceil(n)


# Test the function
sample_size("mean", err_mean=2, sd_mean=10)

# summary.sample_size <- function(object, ...) {
#   if (is.character(object)) {
#     return(object)
#   }

#   cat("Sample size calculation\n")

#   if (object$type == "mean") {
#     cat("Calculation type     : Mean\n")
#     cat("Acceptable Error     :", object$err_mean, "\n")
#     cat("Standard deviation   :", object$sd_mean, "\n")
#   } else {
#     cat("Calculation type     : Proportion\n")
#     cat("Acceptable Error     :", object$err_prop, "\n")
#     cat("Proportion           :", object$p_prop, "\n")
#   }

#   cat("Confidence level     :", object$conf_lev, "\n")
#   cat("Incidence rate       :", object$incidence, "\n")
#   cat("Response rate        :", object$response, "\n")

#   if (object$pop_correction == "no") {
#     cat("Population correction: None\n")
#   } else {
#     cat("Population correction: Yes\n")
#     cat("Population size      :", format_nr(object$pop_size, dec = 0), "\n")
#   }

#   cat("\nRequired sample size     :", format_nr(object$n, dec = 0))
#   cat("\nRequired contact attempts:", format_nr(ceiling(object$n / object$incidence / object$response), dec = 0))

#   rm(object)
# }



__version__ = "0.9.20"
from .bins import *
from .model.perf import *
from .stats import *
from .utils import *
from .props import *
from .model.model import *
from .model.logistic import *
from .model.rforest import *
from .model.mlp import *
from .model.regress import *
from .model.visualize import *
from .notebook import *
from .example_data import *



from starlette.applications import Starlette
from starlette.routing import Mount
from starlette.staticfiles import StaticFiles
from starlette.requests import Request as StarletteRequest
from shiny import App, ui, reactive, Inputs, Outputs, Session
from pathlib import Path
import webbrowser
import nest_asyncio
import uvicorn
import signal
import os
import sys
import tempfile
import pyrsm as rsm
import pyrsm.radiant.utils as ru

# make a radiant_data class with a get_data method
# this can be called from the sub-classes
# https://chat.openai.com/share/0c566cff-6f02-4e55-aec4-e6b379a9aadf
# within each sub-class that method will be made into a reactive.Calc
# check if filters set in the data (?) or radiant_data are passed to the
# sub-classes because dictionaries are mutable ?
# also is the "super" class (data / radiant_data) the place to store state
# from all sub-classes?
# if this works, how to best "rip-out" the data parts of each sub-app and still be
# able to use them individually with a data-filter tab?

# on-session-ended should be used to store information from "input" into the state
# dictionary

# possible to only activate the apps you use? the below suggests you can
# add routes dynamically so ... maybe? ask @joe
# https://github.com/rstudio/py-shiny/issues/352

# see here for how you might be able to use the state dictionary for an
# individual user on a server
# https://discord.com/channels/1109483223987277844/1129448698611511326/1129732244353851432


class data_view:
    def __init__(
        self, datasets: dict, descriptions=None, state=None, code=True, navbar=None
    ) -> None:
        ru.init(
            self,
            datasets,
            descriptions=descriptions,
            state=state,
            code=code,
            navbar=navbar,
        )

    def shiny_ui(self, request: StarletteRequest):
        return ui.page_navbar(
            ru.head_content(),
            ui.nav(
                "Data > View",
                ui.row(
                    ui.column(3, ru.ui_view(self)),
                    ui.column(
                        8,
                        ui.navset_card_tab(
                            ui.nav("View", ru.ui_data_main()),
                        ),
                    ),
                ),
            ),
            self.navbar,
            ru.ui_help(
                "https://github.com/vnijs/pyrsm/blob/main/examples/data/data-view.ipynb",
                "Data > View example notebook",
            ),
            ru.ui_stop(),
            title="Radiant for Python",
            inverse=False,
            id="navbar_id",
        )

    def shiny_server(self, input: Inputs, output: Outputs, session: Session):
        # --- section standard for all apps ---
        ru.make_data_elements(self, input, output, session)

        def update_state():
            with reactive.isolate():
                ru.dct_update(self, input)

        session.on_ended(update_state)

        # --- section standard for all apps ---
        # stops returning code if moved to utils
        @reactive.Effect
        @reactive.event(input.stop, ignore_none=True)
        async def stop_app():
            rsm.md(f"```python\n{self.stop_code}\n```")
            await session.app.stop()
            os.kill(os.getpid(), signal.SIGTERM)


def view(
    data_dct: dict = None,
    descriptions_dct: dict = None,
    state: dict = None,
    code: bool = True,
    host: str = "",
    port: int = 8000,
    log_level: str = "warning",
    debug: bool = False,
):
    """
    Launch a Radiant-for-Python app for single_mean hypothesis testing
    """
    host = ru.set_host(host)
    if data_dct is None:
        data_dct, descriptions_dct = ru.get_dfs(pkg="data")
    rc = data_view(data_dct, descriptions_dct, state=state, code=code)
    nest_asyncio.apply()
    webbrowser.open(f"http://{host}:{port}")
    print(f"Listening on http://{host}:{port}")
    ru.message()

    # redirect stdout and stderr to the temporary file
    if not debug:
        org_stdout = sys.stdout
        org_stderr = sys.stderr
        temp = tempfile.NamedTemporaryFile()
        temp_file = open(temp.name, "w")
        sys.stdout = temp_file
        sys.stderr = temp_file

    app = App(rc.shiny_ui, rc.shiny_server)
    www_dir = Path(__file__).parent.parent.parent / "radiant" / "www"
    app_static = StaticFiles(directory=www_dir, html=False)

    routes = [
        Mount("/www", app=app_static),
        Mount("/", app=app),
    ]

    uvicorn.run(
        Starlette(debug=debug, routes=routes),
        host=host,
        port=port,
        log_level=log_level,
    )

    if not debug:
        sys.stdout = org_stdout
        sys.stderr = org_stderr
        temp_file.close()


if __name__ == "__main__":
    # demand_uk, demand_uk_description = rsm.load_data(pkg="basics", name="demand_uk")
    # data_dct, descriptions_dct = ru.get_dfs(name="demand_uk")
    # single_mean(data_dct, descriptions_dct, code=True)
    view(debug=False)



import pandas as pd
import numpy as np
from pyrsm.model.perf import (
    calc_qnt,
    gains_tab,
    lift_tab,
    lift_plot,
    gains_plot,
    profit_plot,
    ROME_plot,
    evalbin,
)

np.random.seed(1234)
nr = 100
df = pd.DataFrame()
df["x1"] = np.random.uniform(0, 1, nr)
df["x2"] = 1 - df.x1
df["response_prob"] = np.random.uniform(0, 1, nr)
df["response"] = np.where((df["x1"] > 0.5) & (df["response_prob"] > 0.5), "yes", "no")
df["training"] = np.concatenate([np.ones(80), np.zeros(20)])
df["rnd_response"] = np.random.choice(["yes", "no"], nr)


def test_calc_qnt():
    ret = calc_qnt(df, "response", "yes", "x1", qnt=10)
    assert all(
        ret["cum_resp"] == np.array([6, 14, 20, 27, 29, 30, 30, 30, 30, 30])
    ), "Incorrect calculation of cum_resp with x1"


def test_calc_qnt_rev():
    ret = calc_qnt(df, "response", "yes", "x2", qnt=10)
    assert all(
        ret["cum_resp"] == np.array([6, 14, 20, 27, 29, 30, 30, 30, 30, 30])
    ), "Incorrect calculation of cum_resp with x2"


def test_gains_tab():
    ret = gains_tab(df, "response", "yes", "x1", qnt=10)
    assert all(
        ret["cum_gains"].values.round(3)
        == np.array([0.0, 0.2, 0.467, 0.667, 0.9, 0.967, 1.0, 1.0, 1.0, 1.0, 1.0])
    ), "Incorrect calculation of cum_gains"


def test_lift_tab():
    ret = lift_tab(df, "response", "yes", "x1", qnt=10)
    assert all(
        ret["cum_lift"].values.round(3)
        == np.array([2.0, 2.333, 2.222, 2.25, 1.933, 1.667, 1.429, 1.25, 1.111, 1.0])
    ), "Incorrect calculation of cum_lift"


def test_evalbin():
    dct = {"train": df[df.training == 1], "test": df[df.training == 0]}
    ret = evalbin(dct, "response", "yes", "x1", cost=1, margin=2, dec=3)

    assert ret.shape == (2, 18), "Incorrect dimensions"
    assert all(
        ret.loc[0, "profit":"AUC"] == [5, 1.0, 0.116, 0.538, 0.897]
    ), "Errors for profit:AUC in row 0"
    assert all(
        ret.loc[1, "profit":"AUC"] == [-2, 1.0, -0.143, 0.7, 0.845]
    ), "Errors for profit:AUC in row 0"


def test_lift_plot_single():
    fig = lift_plot(df, "response", "yes", "x1", qnt=10)


def test_lift_plot_mult():
    fig = lift_plot(df, "response", "yes", ["x1", "x2"], qnt=10)


def test_gains_plot_single():
    fig = gains_plot(df, "response", "yes", "x1", qnt=10)


def test_gains_plot_mult():
    fig = gains_plot(df, "response", "yes", ["x1", "x2"], qnt=10)


def test_profit_plot_single():
    fig = profit_plot(df, "response", "yes", "x1", qnt=10)


def test_profit_plot_mult():
    fig = profit_plot(df, "response", "yes", ["x1", "x2"], qnt=10, contact=True)


def test_rome_plot_single():
    fig = ROME_plot(df, "response", "yes", "x1", qnt=10)


def test_ROME_plot_mult():
    fig = ROME_plot(df, "response", "yes", ["x1", "x2"], qnt=10)


if __name__ == "__main__":
    test_calc_qnt()
    test_calc_qnt_rev()
    test_gains_tab()
    test_lift_tab()
    test_evalbin()
    test_lift_plot_single()
    test_lift_plot_mult()
    test_gains_plot_single()
    test_gains_plot_mult()
    test_profit_plot_single()
    test_profit_plot_mult()
    test_rome_plot_single()
    test_ROME_plot_mult()



from starlette.applications import Starlette
from starlette.routing import Mount
from starlette.staticfiles import StaticFiles
from starlette.requests import Request as StarletteRequest
from shiny import App, render, ui, reactive, Inputs, Outputs, Session, req
from pathlib import Path
import webbrowser
import nest_asyncio
import uvicorn
import signal
import os
import sys
import tempfile
import pyrsm as rsm
import pyrsm.radiant.utils as ru
import pyrsm.radiant.model.model_utils as mu


def ui_summary(self):
    return (
        ui.panel_conditional(
            "input.tabs == 'Summary'",
            ui.panel_well(
                ui.output_ui("ui_var"),
                ui.input_select(
                    id="alt_hyp",
                    label="Alternative hypothesis:",
                    selected=self.state.get("alt_hyp", "two-sided"),
                    choices={
                        "two-sided": "Two sided",
                        "greater": "Greater than",
                        "less": "Less than",
                    },
                ),
                ui.input_slider(
                    id="conf",
                    label="Confidence level:",
                    min=0,
                    max=1,
                    value=self.state.get("conf", 0.95),
                ),
                ui.input_numeric(
                    id="comp_value",
                    label="Comparison value:",
                    value=self.state.get("comp_value", 0),
                ),
            ),
        ),
    )


choices = {
    # "None": "None",
    "hist": "Histogram",
    # "sim": "Simulate",
}


class basics_single_mean:
    def __init__(
        self, datasets: dict, descriptions=None, state=None, code=True, navbar=None
    ) -> None:
        ru.init(
            self,
            datasets,
            descriptions=descriptions,
            state=state,
            code=code,
            navbar=navbar,
        )

    def shiny_ui(self, request: StarletteRequest):
        return ui.page_navbar(
            ru.head_content(),
            ui.nav(
                "Basics > Single mean",
                ui.row(
                    ui.column(
                        3,
                        ru.ui_data(self),
                        ui_summary(self),
                        ru.ui_plot(self, choices),
                    ),
                    ui.column(8, ru.ui_main_basics()),
                ),
            ),
            self.navbar,
            ru.ui_help(
                "https://github.com/vnijs/pyrsm/blob/main/examples/basics/basics-single-mean.ipynb",
                "Single mean example notebook",
            ),
            ru.ui_stop(),
            title="Radiant for Python",
            inverse=False,
            id="navbar_id",
        )

    def shiny_server(self, input: Inputs, output: Outputs, session: Session):
        # --- section standard for all apps ---
        get_data = ru.make_data_elements(self, input, output, session)

        def update_state():
            with reactive.isolate():
                ru.dct_update(self, input)

        session.on_ended(update_state)

        # --- section unique to each app ---
        @output(id="ui_var")
        @render.ui
        def ui_var():
            isNum = get_data()["var_types"]["isNum"]
            return ui.input_select(
                id="var",
                label="Variable (select one)",
                selected=self.state.get("var", None),
                choices=isNum,
            )

        def estimation_code():
            data_name, code = (get_data()[k] for k in ["data_name", "code"])
            req(input.comp_value() is not None, "comp_value must be specified")

            args = {
                "data": f"""{{"{data_name}": {data_name}}}""",
                "var": input.var(),
                "alt_hyp": input.alt_hyp(),
                "conf": input.conf(),
                "comp_value": input.comp_value(),
            }

            args_string = ru.drop_default_args(args, rsm.basics.single_mean)
            return f"""rsm.basics.single_mean({args_string})""", code

        show_code, estimate = mu.make_estimate(
            self,
            input,
            output,
            get_data,
            fun="basics.single_mean",
            ret="sm",
            ec=estimation_code,
            run=False,
            debug=True,
        )

        def summary_code():
            return """sm.summary()"""

        mu.make_summary(
            self,
            input,
            output,
            session,
            show_code,
            estimate,
            ret="sm",
            sum_fun=rsm.basics.single_mean.summary,
            sc=summary_code,
        )

        mu.make_plot(
            self,
            input,
            output,
            session,
            show_code,
            estimate,
            ret="sm",
        )

        # --- section standard for all apps ---
        # stops returning code if moved to utils
        @reactive.Effect
        @reactive.event(input.stop, ignore_none=True)
        async def stop_app():
            rsm.md(f"```python\n{self.stop_code}\n```")
            await session.app.stop()
            os.kill(os.getpid(), signal.SIGTERM)


def single_mean(
    data_dct: dict = None,
    descriptions_dct: dict = None,
    state: dict = None,
    code: bool = True,
    host: str = "",
    port: int = 8000,
    log_level: str = "warning",
    debug: bool = False,
):
    """
    Launch a Radiant-for-Python app for single_mean hypothesis testing
    """
    host = ru.set_host(host)
    if data_dct is None:
        data_dct, descriptions_dct = ru.get_dfs(pkg="basics", name="demand_uk")
    rc = basics_single_mean(data_dct, descriptions_dct, state=state, code=code)
    nest_asyncio.apply()
    webbrowser.open(f"http://{host}:{port}")
    print(f"Listening on http://{host}:{port}")
    ru.message()

    # redirect stdout and stderr to the temporary file
    if not debug:
        org_stdout = sys.stdout
        org_stderr = sys.stderr
        temp = tempfile.NamedTemporaryFile()
        temp_file = open(temp.name, "w")
        sys.stdout = temp_file
        sys.stderr = temp_file

    app = App(rc.shiny_ui, rc.shiny_server)
    www_dir = Path(__file__).parent.parent.parent / "radiant" / "www"
    app_static = StaticFiles(directory=www_dir, html=False)

    routes = [
        Mount("/www", app=app_static),
        Mount("/", app=app),
    ]

    uvicorn.run(
        Starlette(debug=debug, routes=routes),
        host=host,
        port=port,
        log_level=log_level,
    )

    if not debug:
        sys.stdout = org_stdout
        sys.stderr = org_stderr
        temp_file.close()


if __name__ == "__main__":
    # demand_uk, demand_uk_description = rsm.load_data(pkg="basics", name="demand_uk")
    # data_dct, descriptions_dct = ru.get_dfs(name="demand_uk")
    # single_mean(data_dct, descriptions_dct, code=True)
    single_mean(debug=False)



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import polars as pl
from scipy import stats
from pyrsm.model import sig_stars
from pyrsm.utils import ifelse, check_dataframe
import pyrsm.basics.utils as bu
from typing import Union


class single_mean:
    def __init__(
        self,
        data: pd.DataFrame | pl.DataFrame | dict[str, pd.DataFrame | pl.DataFrame],
        var: str,
        alt_hyp: str = "two-sided",
        conf: float = 0.95,
        comp_value: float = 0,
    ):
        if isinstance(data, dict):
            self.name = list(data.keys())[0]
            self.data = data[self.name].copy()
        else:
            self.data = data
            self.name = "Not provided"

        self.data = check_dataframe(self.data)
        self.var = var
        self.alt_hyp = alt_hyp
        self.conf = conf
        self.comp_value = comp_value

        result = stats.ttest_1samp(
            a=self.data[self.var],
            popmean=self.comp_value,
            nan_policy="omit",
            alternative=self.alt_hyp,
        )

        self.t_val, self.p_val = result.statistic, result.pvalue
        self.ci = result.confidence_interval(confidence_level=conf)

        self.mean = np.nanmean(self.data[self.var])
        self.n = len(self.data[self.var])
        self.n_missing = self.data[self.var].isna().sum()

        self.sd = self.data[self.var].std()
        self.se = self.data[self.var].sem()
        tscore = stats.t.ppf((1 + self.conf) / 2, self.n - 1)

        self.me = (tscore * self.se).real
        self.diff = self.mean - self.comp_value
        self.df = self.n - 1

    def summary(self, dec=3) -> None:
        print("Single mean test")
        print(f"Data      : {self.name}")
        print(f"Variables : {self.var}")
        print(f"Confidence: {self.conf}")
        print(f"Comparison: {self.comp_value}\n")
        print(f"Null hyp. : the mean of {self.var} is equal to {self.comp_value}")

        if self.alt_hyp == "less":
            alt_hyp = "less than"
        elif self.alt_hyp == "two-sided":
            alt_hyp = "not equal to"
        else:
            alt_hyp = "greater than"

        cl = bu.ci_label(self.alt_hyp, self.conf, dec=dec)

        print(f"Alt. hyp. : the mean of {self.var} is {alt_hyp} {self.comp_value}\n")

        row1 = [[self.mean, self.n, self.n_missing, self.sd, self.se, self.me]]
        row2 = [
            [
                self.diff,
                self.se,
                self.t_val,
                ifelse(self.p_val < 0.001, "< .001", self.p_val),
                self.df,
                self.ci[0],
                self.ci[1],
                sig_stars([self.p_val])[0],
            ]
        ]

        col_names1 = ["mean", "n", "n_missing", "sd", "se", "me"]
        col_names2 = ["diff", "se", "t.value", "p.value", "df", cl[0], cl[1], ""]

        table1 = pd.DataFrame(row1, columns=col_names1).round(dec)
        table2 = pd.DataFrame(row2, columns=col_names2).round(dec)

        print(table1.to_string(index=False))
        print(table2.to_string(index=False))
        print("\nSignif. codes:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1")

    def plot(self, plots: str = "hist") -> None:
        if plots == "hist":
            fig = self.data[self.var].plot.hist(title=self.var, color="slateblue")
            plt.vlines(
                x=(self.comp_value, self.ci[0], self.mean, self.ci[1]),
                ymin=fig.get_ylim()[0],
                ymax=fig.get_ylim()[1],
                colors=("r", "k", "k", "k"),
                linestyles=("solid", "dashed", "solid", "dashed"),
            )
        elif plots == "sim":
            print("Plot type not available yet")
            # self.data[self.var].plot.hist(title=self.var, color="slateblue")
            # plt.vlines(
            #     x=[self.comp_value, self.me, self.mean],
            #     colors=["r", "k", "k"],
            #     linestyles=["solid", "dashed", "dashed"],
            # )
        else:
            print("Invalid plot type")



from starlette.applications import Starlette
from starlette.routing import Mount
from starlette.staticfiles import StaticFiles
from starlette.requests import Request as StarletteRequest
from pathlib import Path
from shiny import App, render, ui, reactive, Inputs, Outputs, Session
import webbrowser
import nest_asyncio
import uvicorn
import signal
import os
import sys
import tempfile
import pyrsm as rsm
import pyrsm.radiant.utils as ru
import pyrsm.radiant.model.model_utils as mu

choices = {
    "observed": "Observed",
    "expected": "Expected",
    "chisq": "Chi-squared",
    "dev_std": "Deviation std.",
    "perc_row": "Row percentages",
    "perc_col": "Column percentages",
    "perc": "Table percentages",
}


def ui_summary(self):
    return ui.panel_conditional(
        "input.tabs == 'Summary'",
        ui.panel_well(
            ui.output_ui("ui_var1"),
            ui.output_ui("ui_var2"),
            ui.input_checkbox_group(
                id="output",
                label="Select output tables:",
                selected=self.state.get("output", None),
                choices=choices,
            ),
        ),
    )


plots = {"None": "None"}
plots.update(choices)


class basics_cross_tabs:
    def __init__(
        self, datasets: dict, descriptions=None, state=None, code=True, navbar=None
    ) -> None:
        ru.init(
            self,
            datasets,
            descriptions=descriptions,
            state=state,
            code=code,
            navbar=navbar,
        )

    def shiny_ui(self, request: StarletteRequest):
        return ui.page_navbar(
            ru.head_content(),
            ui.nav(
                "Basics > Cross-tabs",
                ui.row(
                    ui.column(
                        3,
                        ru.ui_data(self),
                        ui_summary(self),
                        ru.ui_plot(self, plots),
                    ),
                    ui.column(8, ru.ui_main_basics(self)),
                ),
            ),
            self.navbar,
            ru.ui_help(
                "https://github.com/vnijs/pyrsm/blob/main/examples/basics/basics-cross-tabs.ipynb",
                "Cross-tabs example notebook",
            ),
            ru.ui_stop(),
            title="Radiant for Python",
            inverse=False,
            id="navbar_id",
        )

    def shiny_server(self, input: Inputs, output: Outputs, session: Session):
        # --- section standard for all apps ---
        get_data = ru.make_data_elements(self, input, output, session)

        def update_state():
            with reactive.isolate():
                ru.dct_update(self, input)

        session.on_ended(update_state)

        def update_state():
            with reactive.isolate():
                avoid = ["copy_reset"]  # because async?
                input_keys = [
                    k
                    for k in input.__dict__["_map"].keys()
                    if k[0] != "." and k not in avoid
                ]
                # print(type(input.copy_reset())) # can't even run this without an error
                self.state.update({k: input[k]() for k in input_keys})
                self.state.update(
                    {k: list(v) for k, v in self.state.items() if isinstance(v, tuple)}
                )

        session.on_ended(update_state)

        # --- section unique to each app ---
        @output(id="ui_var1")
        @render.ui
        def ui_var1():
            isCat = get_data()["var_types"]["isCat"]
            return ui.input_select(
                id="var1",
                label="Select a categorical variable:",
                selected=self.state.get("var1", None),
                choices=isCat,
            )

        @output(id="ui_var2")
        @render.ui
        def ui_var2():
            isCat = get_data()["var_types"]["isCat"].copy()
            if (input.var1() is not None) and (input.var1() in isCat):
                del isCat[input.var1()]

            return ui.input_select(
                id="var2",
                label="Select a categorical variable:",
                selected=self.state.get("var2", None),
                choices=isCat,
            )

        def estimation_code():
            data_name, code = (get_data()[k] for k in ["data_name", "code"])

            args = {
                "data": f"""{{"{data_name}": {data_name}}}""",
                "var1": input.var1(),
                "var2": input.var2(),
            }

            args_string = ru.drop_default_args(args, rsm.basics.cross_tabs)
            return f"""rsm.basics.cross_tabs({args_string})""", code

        show_code, estimate = mu.make_estimate(
            self,
            input,
            output,
            get_data,
            fun="basics.cross_tabs",
            ret="ct",
            ec=estimation_code,
            run=False,
            debug=True,
        )

        def summary_code():
            args = [c for c in input.output()]
            return f"""ct.summary(output={args})"""

        mu.make_summary(
            self,
            input,
            output,
            session,
            show_code,
            estimate,
            ret="ct",
            sum_fun=rsm.basics.cross_tabs.summary,
            sc=summary_code,
        )

        def plot_code():
            return f"""ct.plot(plots="{input.plots()}")"""

        mu.make_plot(
            self,
            input,
            output,
            session,
            show_code,
            estimate,
            ret="ct",
            pc=plot_code,
        )

        # --- section standard for all apps ---
        # stops returning code if moved to utils
        @reactive.Effect
        @reactive.event(input.stop, ignore_none=True)
        async def stop_app():
            rsm.md(f"```python\n{self.stop_code}\n```")
            await session.app.stop()
            os.kill(os.getpid(), signal.SIGTERM)


def cross_tabs(
    data_dct: dict = None,
    descriptions_dct: dict = None,
    state: dict = None,
    code: bool = True,
    host: str = "",
    port: int = 8000,
    log_level: str = "warning",
    debug: bool = False,
):
    """
    Launch a Radiant-for-Python app for cross-tabs hypothesis testing
    """
    host = ru.set_host(host)
    if data_dct is None:
        data_dct, descriptions_dct = ru.get_dfs(pkg="basics", name="newspaper")
    rc = basics_cross_tabs(data_dct, descriptions_dct, state=state, code=code)
    nest_asyncio.apply()
    webbrowser.open(f"http://{host}:{port}")
    print(f"Listening on http://{host}:{port}")
    ru.message()

    # redirect stdout and stderr to the temporary file
    if not debug:
        org_stdout = sys.stdout
        org_stderr = sys.stderr
        temp = tempfile.NamedTemporaryFile()
        temp_file = open(temp.name, "w")
        sys.stdout = temp_file
        sys.stderr = temp_file

    app = App(rc.shiny_ui, rc.shiny_server)
    www_dir = Path(__file__).parent.parent.parent / "radiant" / "www"
    app_static = StaticFiles(directory=www_dir, html=False)

    routes = [
        Mount("/www", app=app_static),
        Mount("/", app=app),
    ]

    uvicorn.run(
        Starlette(debug=debug, routes=routes),
        host=host,
        port=port,
        log_level=log_level,
    )

    if not debug:
        sys.stdout = org_stdout
        sys.stderr = org_stderr
        temp_file.close()


if __name__ == "__main__":
    cross_tabs(debug=False)



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import polars as pl
from scipy import stats
import seaborn as sns
from pyrsm.model.model import sig_stars
from pyrsm.utils import check_dataframe
from typing import Optional, Union


class correlation:
    """
    Calculate correlations between numeric variables in a Pandas dataframe

    Parameters
    ----------
    data : Pandas dataframe with numeric variables

    Returns
    -------
    Correlation object with two key attributes
    cr: Correlation matrix
    cp: p.value matrix
    cv: Covariance matrix

    Examples
    --------
    import pandas as pd
    import pyrsm as rsm
    salary, salary_description = rsm.load_data(pkg="basics", name="salary")
    cr = rsm.correlation(salary[["salary", "yrs.since.phd", "yrs.service"]])
    cr.cr
    """

    def __init__(
        self,
        data: pd.DataFrame | pl.DataFrame | dict[str, pd.DataFrame | pl.DataFrame],
        vars: Optional[list[str]] = [],
        method: str = "pearson",
    ) -> None:
        if isinstance(data, dict):
            self.name = list(data.keys())[0]
            self.data = data[self.name].copy()
        else:
            self.data = data
            self.name = "Not provided"

        self.data = check_dataframe(self.data)
        self.vars = vars
        if len(self.vars) == 0:
            self.vars = [
                col
                for col in self.data.columns
                if pd.api.types.is_numeric_dtype(self.data[col].dtype)
            ]
        self.data = self.data[self.vars].copy()
        self.method = method

        ncol = self.data.shape[1]
        cr = np.zeros([ncol, ncol])
        cp = cr.copy()
        cv = cr.copy()
        for i in range(ncol - 1):
            for j in range(i + 1, ncol):
                cdf = self.data.iloc[:, [i, j]]
                # pairwise deletion
                cdf = cdf[~np.any(np.isnan(cdf), axis=1)]
                if self.method == "spearman":
                    c = stats.spearmanr(cdf.iloc[:, 0], cdf.iloc[:, 1])
                elif self.method == "kendall":
                    c = stats.kendalltau(cdf.iloc[:, 0], cdf.iloc[:, 1])
                else:
                    c = stats.pearsonr(cdf.iloc[:, 0], cdf.iloc[:, 1])

                cr[j, i] = c[0]
                cp[j, i] = c[1]
                cv[j, i] = cdf.iloc[:, [0, 1]].cov().iloc[0, 1]

        self.cr = cr
        self.cp = cp
        self.cv = cv

    def summary(self, cov=False, cutoff: float = 0, dec: int = 2) -> None:
        """
        Print correlations between numeric variables in a Pandas dataframe

        Parameters
        ----------
        cov : bool
            Show the covariance matrix if set to True
        cutoff : float
            Only show correlations larger than a threshold in absolute value
        dec : int
            Number of decimal places to use in rounding

        Examples
        --------
        import pandas as pd
        import pyrsm as rsm
        salary, salary_description = load_data(pkg="basics", name="salary")
        cr = rsm.correlation(salary[["salary", "yrs.since.phd", "yrs.service"]])
        cr.summary()
        """
        prn = "Correlation\n"
        prn += f"Data     : {self.name}\n"
        prn += f"Method   : {self.method}"
        print(prn)
        if len(self.vars) < 2:
            print("\n**Select two or more variables to calculate correlations**")
        else:
            ind = np.triu_indices(self.cr.shape[0])
            cn = list(self.data.columns)

            # correlations
            crs = pd.DataFrame(self.cr.round(dec).astype(str), columns=cn, index=cn)
            if cutoff > 0:
                crs.values[np.abs(self.cr) < cutoff] = ""
            crs.values[ind] = ""

            # p.values
            cps = pd.DataFrame(self.cp.round(dec).astype(str), columns=cn, index=cn)
            if cutoff > 0:
                cps.values[np.abs(self.cr) < cutoff] = ""
            cps.values[ind] = ""

            if len(cn) > 2:
                x = "x"
                y = "y"
            else:
                x = cn[0]
                y = cn[1]

            prn = f"Cutoff   : {cutoff}\n"
            prn += "Variables: " + ", ".join(cn) + "\n"
            prn += f"Null hyp.: variables {x} and {y} are not correlated\n"
            prn += f"Alt. hyp.: variables {x} and {y} are correlated\n"
            print(prn)
            print("Correlation matrix:")
            print(crs.iloc[1:, :-1])
            print("\np.values:")
            print(cps.iloc[1:, :-1])

            if cov:
                cvs = pd.DataFrame(self.cv.round(dec), columns=cn, index=cn).map(
                    lambda x: "{:,}".format(x)
                )
                if cutoff > 0:
                    cvs.values[np.abs(self.cr) < cutoff] = ""

                cvs.values[ind[0], ind[1]] = ""
                print("\nCovariance matrix:")
                print(cvs.iloc[1:, :-1])

    def plot(self, nobs: int = 1000, dec: int = 2, figsize: tuple[float, float] = None):
        """
        Plot of correlations between numeric variables in a Pandas dataframe

        Parameters
        ----------
        nobs : int
            Number of observations to use for the scatter plots. The default
            value is 1,000. To use all observations in the plots, use nobs=-1
        dec : int
            Number of decimal places to use in rounding
        figsize : tuple
            A tuple that determines the figure size. If None, size is
            determined based on the number of numeric variables in the
            data

        Examples
        --------
        import pandas as pd
        import pyrsm as rsm
        rsm.(pkg="basics", name="salary", dct=globals())
        cr = rsm.correlation(salary[["salary", "yrs.since.phd", "yrs.service"]])
        cr.plot(figsize=(7, 3))
        """

        if figsize is None:
            figsize = (max(5, self.cr.shape[0]), max(self.cr.shape[0], 5))

        def cor_label(label, longest, ax_sub):
            ax_sub.axes.xaxis.set_visible(False)
            ax_sub.axes.yaxis.set_visible(False)
            # set font size to avoid exceeding boundaries
            fs = min(figsize[0], figsize[1])
            font = (80 * fs) / (len(longest) * self.cr.shape[0])
            ax_sub.text(
                0.5,
                0.5,
                label,
                horizontalalignment="center",
                verticalalignment="center",
                fontsize=font,
            )

        def cor_text(r, p, ax_sub, dec=2):
            if np.isnan(p):
                p = 1

            p = round(p, dec)
            rt = round(r, dec)
            p_star = sig_stars([p])[0]

            fs = min(figsize[0], figsize[1])
            font = (50 * fs) / (len(str(rt)) * len(self.data.columns))
            font_star = (12 * fs) / len(self.data.columns)

            ax_sub.axes.xaxis.set_visible(False)
            ax_sub.axes.yaxis.set_visible(False)
            ax_sub.text(
                0.5,
                0.5,
                rt,
                horizontalalignment="center",
                verticalalignment="center",
                fontsize=font * abs(r),
            )
            ax_sub.text(
                0.8,
                0.8,
                p_star,
                horizontalalignment="center",
                verticalalignment="center",
                fontsize=font_star,
                color="blue",
            )

        def cor_plot(x_data, y_data, ax_sub, s_size):
            sns.regplot(
                x=x_data,
                y=y_data,
                ax=ax_sub,
                scatter_kws={"alpha": 0.3, "color": "black", "s": s_size},
                line_kws={"color": "blue"},
            )

            ax_sub.axes.xaxis.set_visible(False)
            ax_sub.axes.yaxis.set_visible(False)

        def cor_mat(data, cmat, pmat, dec=2, nobs=1000, figsize=None):
            cn = data.columns
            ncol = len(cn)
            longest = max(cn, key=len)
            s_size = 50 / len(self.data.columns)

            fs = min(figsize[0], figsize[1])
            s_size = (5 * fs) / len(self.data.columns)

            _, axes = plt.subplots(ncol, ncol, figsize=figsize)

            if nobs < data.shape[0] and nobs != np.Inf and nobs != -1:
                data = data.copy().sample(nobs)

            for i in range(ncol):
                for j in range(ncol):
                    if i == j:
                        cor_label(cn[i], longest, axes[i, j])
                    elif i > j:
                        cor_plot(data[cn[i]], data[cn[j]], axes[i, j], s_size)
                    else:
                        cor_text(cmat[j, i], pmat[j, i], axes[i, j], dec=2)

            plt.subplots_adjust(wspace=0.04, hspace=0.04)

        cor_mat(self.data, self.cr, self.cp, dec=dec, nobs=nobs, figsize=figsize)



import pickle


def save_state(dct, path):
    """
    Store the (partial) state of a Jupyter notebook

    Parameters
    ----------
    dct : dict
        Dictionary of python objects to store (e.g., globals())
    path : str
        File path to use to store state. Use .state.pkl as
        the file name extension

    Examples
    --------
    save_state(globals(), path="my-notebook.state.pkl")
    """

    remove_keys = [
        "In",
        "Out",
        "get_ipython",
        "exit",
        "quit",
        "json",
        "sys",
        "NamespaceMagics",
        "state",
        "remove_keys",
        "remove_types",
    ]
    remove_types = [
        "<class 'module'>",
        "<class 'function'>",
        "<class 'builtin_function_or_method'>",
        "<class 'abc.ABCMeta'>",
        "<class 'type'>",
        "<class '_io.BufferedReader'>",
        "<class 'weakref'>",
    ]
    state = {
        key: val
        for key, val in dct.items()
        if (
            not key.startswith("_")
            and (key not in remove_keys)
            and (str(type(val)) not in remove_types)
        )
    }
    try:
        with open(path, "wb") as f:
            pickle.dump(state, f)
    except Exception as err:
        print(err)
        print("\nSome objects in globals() could not be 'pickled'")
        print("Either remove those objects or explicitly pass in")
        print("a dictionary with the objects you want to store")


def load_state(path, dct=None):
    """
    Re-store the (partial) state of a Jupyter notebook

    Parameters
    ----------
    path : str
        Path to state file location
    dct : dict
        Dictionary to add python objects to (e.g., globals()).
        If None, a dictionary of objects will be returned

    Examples
    --------
    load_state("my-notebook.state.pkl", globals())
    """

    try:
        with open(path, "rb") as f:
            g = pickle.load(f)

        if dct is None:
            return g
        else:
            # using the mutability feature in python
            for key, val in g.items():
                dct[key] = val
    except Exception as err:
        print(err)
        print("\nCould not load file: " + path)



from typing import Union, Optional
import pandas as pd
import polars as pl
import numpy as np
from statsmodels.genmod.families import Binomial
from statsmodels.genmod.families.links import Logit
import statsmodels.formula.api as smf
from scipy import stats
from pyrsm.utils import ifelse, setdiff, check_dataframe
from pyrsm.model.model import (
    sig_stars,
    model_fit,
    or_ci,
    or_plot,
    sim_prediction,
    predict_ci,
    convert_binary,
    convert_to_list,
)
from pyrsm.model.model import vif as calc_vif
from .visualize import pred_plot_sm, vimp_plot_sm, extract_evars, extract_rvar


class logistic:
    """
    Initialize logistic regression model

    Parameters
    ----------
    data: pandas DataFrame; dataset
    rvar: String; name of the column to be used as the response variable
    lev: String; name of the level in the response variable
    evar: List of strings; contains the names of the columns of data to be used as explanatory variables
    ivar: List of strings; contains the names of columns to interact and add as explanatory variables (e.g., ["x1:x2", "x3:x4])
    form: String; formula for the regression equation to use if evar and rvar are not provided
    weights: String; name of the column that contains frequency weights
    """

    def __init__(
        self,
        data: pd.DataFrame | pl.DataFrame | dict[str, pd.DataFrame | pl.DataFrame],
        rvar: Optional[str] = None,
        lev: Optional[str] = None,
        evar: Optional[list[str]] = None,
        ivar: Optional[list[str]] = None,
        form: Optional[str] = None,
        weights: Optional[str] = None,
    ) -> None:
        if isinstance(data, dict):
            self.name = list(data.keys())[0]
            self.data = data[self.name]
        else:
            self.data = data
            self.name = "Not provided"
        self.data = check_dataframe(self.data)
        self.rvar = rvar
        self.lev = lev
        self.evar = convert_to_list(evar)
        self.ivar = convert_to_list(ivar)
        self.form = form
        if weights is not None and weights != "None":
            self.weights_name = weights
            self.weights = self.data[weights]
        else:
            self.weights_name = self.weights = None

        if self.lev is not None and self.rvar is not None:
            self.data[self.rvar] = convert_binary(self.data[self.rvar], self.lev)

        if self.form:
            self.fitted = smf.glm(
                formula=self.form,
                data=self.data,
                freq_weights=self.weights,
                family=Binomial(link=Logit()),
            ).fit()
            self.evar = extract_evars(self.fitted.model, self.data.columns)
            self.rvar = extract_rvar(self.fitted.model, self.data.columns)
            if self.lev is None:
                self.lev = self.data.at[0, self.rvar]
        else:
            if self.evar is None or len(self.evar) == 0:
                self.form = f"{self.rvar} ~ 1"
            else:
                self.form = f"{self.rvar} ~ {' + '.join(self.evar)}"
            if self.ivar:
                self.form += f" + {' + '.join(self.ivar)}"
            self.fitted = smf.glm(
                formula=self.form,
                data=self.data,
                freq_weights=self.weights,
                family=Binomial(link=Logit()),
            ).fit()
        df = pd.DataFrame(np.exp(self.fitted.params), columns=["OR"]).dropna()
        df["OR%"] = 100 * ifelse(df["OR"] < 1, -(1 - df["OR"]), df["OR"] - 1)
        df["coefficient"] = self.fitted.params
        df["std.error"] = self.fitted.params / self.fitted.tvalues
        # wierd but this is what statsmodels uses in summary
        df["z.value"] = self.fitted.tvalues
        df["p.value"] = self.fitted.pvalues
        df["  "] = sig_stars(self.fitted.pvalues)
        self.coef = df.reset_index()

    def summary(
        self, main=True, fit=True, ci=False, vif=False, test=None, dec=3
    ) -> None:
        """
        Summarize output from a logistic regression model

        """
        if main:
            print("Logistic regression (GLM)")
            print(f"Data                 : {self.name}")
            print(f"Response variable    : {self.rvar}")
            print(f"Level                : {self.lev}")
            print(f"Explanatory variables: {', '.join(self.evar)}")
            if self.weights is not None:
                print(f"Weights used         : {self.weights_name}")
            print(f"Null hyp.: There is no effect of x on {self.rvar}")
            print(f"Alt. hyp.: There is an effect of x on {self.rvar}")

            df = self.coef.copy()
            df["OR"] = df["OR"].round(dec)
            df["coefficient"] = df["coefficient"].round(2)
            df["std.error"] = df["std.error"].round(dec)
            df["z.value"] = df["z.value"].round(dec)
            df["p.value"] = ifelse(
                df["p.value"] < 0.001, "< .001", df["p.value"].round(dec)
            )
            df["OR%"] = [f"{round(o, max(dec-2, 0))}%" for o in df["OR%"]]
            df["index"] = df["index"].str.replace("[T.", "[", regex=False)
            df = df.set_index("index")
            df.index.name = None
            print(f"\n{df.to_string()}")
            print("\nSignif. codes:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1")

        if fit:
            print(f"\n{model_fit(self.fitted, dec=dec)}")

        if ci:
            print("\nConfidence intervals:")
            df = or_ci(self.fitted, dec=dec).set_index("index")
            df.index.name = None
            print(f"\n{df.to_string()}")

        if vif:
            if self.evar is None or len(self.evar) < 2:
                print("\nVariance Inflation Factors cannot be calculated")
            else:
                print("\nVariance inflation factors:")
                print(f"\n{calc_vif(self.fitted, dec=dec).to_string()}")

        if test is not None and len(test) > 0:
            self.chisq_test(test=test, dec=dec)

    def predict(
        self, data=None, cmd=None, data_cmd=None, ci=False, conf=0.95
    ) -> pd.DataFrame:
        """
        Predict probabilities for a logistic regression model
        """
        if data is None:
            data = self.data
        data = data.loc[:, self.evar].copy()
        if data_cmd is not None:
            for k, v in data_cmd.items():
                data[k] = v
        elif cmd is not None:
            cmd = {k: ifelse(isinstance(v, str), [v], v) for k, v in cmd.items()}
            data = sim_prediction(data=data, vary=cmd)

        if ci:
            if data_cmd is not None:
                raise ValueError(
                    "Confidence intervals not available when using the Data & Command option"
                )
            else:
                return pd.concat(
                    [data, predict_ci(self.fitted, data, conf=conf)], axis=1
                )
        else:
            return data.assign(prediction=self.fitted.predict(data))

    def plot(
        self,
        plots="or",
        data=None,
        alpha=0.05,
        intercept=False,
        incl=None,
        excl=None,
        incl_int=[],
        fix=True,
        hline=False,
        nnv=20,
        minq=0.025,
        maxq=0.975,
        figsize=None,
    ) -> None:
        """
        Plots for a logistic regression model
        """
        if "or" in plots:
            or_plot(
                self.fitted,
                alpha=alpha,
                intercept=intercept,
                incl=incl,
                excl=excl,
                figsize=figsize,
            )
        if "pred" in plots:
            pred_plot_sm(
                self.fitted,
                data=ifelse(data is None, self.data, data),
                incl=incl,
                excl=ifelse(excl is None, [], excl),
                incl_int=incl_int,
                fix=fix,
                hline=hline,
                nnv=nnv,
                minq=minq,
                maxq=maxq,
            )
        if "vimp" in plots:
            vimp_plot_sm(
                self.fitted,
                data=ifelse(data is None, self.data, data),
                rep=10,
                ax=None,
                ret=False,
            )

    def chisq_test(self, test=None, dec=3) -> None:
        """
        Chisq-test for competing models

        Parameters
        ----------
        test : list
            List of strings; contains the names of the columns of data to be tested
        """
        if test is None:
            test = self.evar
        else:
            test = ifelse(isinstance(test, str), [test], test)

        evar = [c for c in self.evar if c not in test]
        if self.ivar is not None and len(self.ivar) > 0:
            sint = setdiff(self.ivar, test)
            test += [s for t in test for s in sint if f"I({t}" not in s and t in s]
            sint = setdiff(sint, test)
        else:
            sint = []

        form = f"{self.rvar} ~ "
        if len(evar) == 0 and len(sint) == 0:
            form += "1"
        else:
            form += f"{' + '.join(evar + sint)}"

        # ensure constraints are unique
        # pattern = r"(\[T\.[^\]]*\])\:"
        # hypotheses = list(
        #     set(
        #         [
        #             f"({c} = 0)"
        #             for c in self.fitted.model.exog_names
        #             for v in test
        #             if f"{v}:" in c
        #             or f":{v}" in c
        #             or f"{v}[T." in c
        #             or v == c
        #             or v == re.sub(pattern, ":", c)
        #         ]
        #     )
        # )

        print(f"\nModel 1: {form}")
        print(f"Model 2: {self.form}")

        # Wald test (faster but not as accurate)
        # out = self.fitted.wald_test(hypotheses, scalar=True)
        # pvalue = ifelse(out.pvalue < 0.001, "< .001", round(out.pvalue, dec))
        # print(
        #     f"Chi-squared: {round(out.statistic, dec)} df ({out.df_denom:.0f}), p.value {pvalue}"
        # )

        # LR test of competing models (slower but more accurate)
        sub_fitted = smf.glm(
            formula=form,
            data=self.data,
            freq_weights=self.weights,
            family=Binomial(link=Logit()),
        ).fit()

        lrtest = -2 * (sub_fitted.llf - self.fitted.llf)
        df = self.fitted.df_model - sub_fitted.df_model
        pvalue = stats.chi2.sf(lrtest, df)

        # calculate pseudo R-squared values for both models
        pr2_full = 1 - self.fitted.llf / self.fitted.llnull
        pr2_sub = 1 - sub_fitted.llf / sub_fitted.llnull

        print(f"Pseudo R-squared, Model 1 vs 2: {pr2_sub:.3f} vs {pr2_full:.3f}")
        pvalue = ifelse(pvalue < 0.001, "< .001", round(pvalue, dec))
        print(f"Chi-squared: {round(lrtest, dec)} df ({df:.0f}), p.value {pvalue}")



import pandas as pd
import polars as pl
import re
import statsmodels.formula.api as smf
from typing import Optional, Union
from pyrsm.utils import ifelse, format_nr, setdiff, check_dataframe
from pyrsm.model.visualize import pred_plot_sm, vimp_plot_sm
from pyrsm.model.model import (
    sig_stars,
    model_fit,
    extract_evars,
    extract_rvar,
    scatter_plot,
    reg_dashboard,
    residual_plot,
    coef_plot,
    coef_ci,
    predict_ci,
    sim_prediction,
    convert_to_list,
)
from pyrsm.model.model import vif as calc_vif
from pyrsm.model.visualize import distr_plot
from pyrsm.basics.correlation import correlation


class regress:
    """
    Estimate linear regression model

    Parameters
    ----------
    data: pandas DataFrame; dataset
    evar: List of strings; contains the names of the columns of data to be used as explanatory variables
    rvar: String; name of the column to be used as the response variable
    ivar: List of strings; contains the names of columns to interact and add as explanatory variables (e.g., ["x1:x2", "x3:x4])
    form: String; formula for the regression equation to use if evar and rvar are not provided
    """

    def __init__(
        self,
        data: pd.DataFrame | pl.DataFrame | dict[str, pd.DataFrame | pl.DataFrame],
        rvar: Optional[str] = None,
        evar: Optional[list[str]] = None,
        ivar: Optional[list[str]] = None,
        form: Optional[str] = None,
    ) -> None:
        if isinstance(data, dict):
            self.name = list(data.keys())[0]
            self.data = data[self.name].copy()  # needed with pandas
        else:
            self.data = data  # needed with pandas
            self.name = "Not provided"
        self.data = check_dataframe(self.data)
        self.rvar = rvar
        self.evar = convert_to_list(evar)
        self.ivar = convert_to_list(ivar)
        self.form = form

        if self.form:
            self.fitted = smf.ols(formula=self.form, data=self.data).fit()
            self.evar = extract_evars(self.fitted.model, self.data.columns)
            self.rvar = extract_rvar(self.fitted.model, self.data.columns)
        else:
            if self.evar is None or len(self.evar) == 0:
                self.form = f"{self.rvar} ~ 1"
            else:
                self.form = f"{self.rvar} ~ {' + '.join(self.evar)}"
            if self.ivar:
                self.form += f" + {' + '.join(self.ivar)}"
            self.fitted = smf.ols(self.form, data=self.data).fit()

        df = pd.DataFrame(self.fitted.params, columns=["coefficient"]).dropna()
        df["std.error"] = self.fitted.params / self.fitted.tvalues
        df["t.value"] = self.fitted.tvalues
        df["p.value"] = self.fitted.pvalues
        df["  "] = sig_stars(self.fitted.pvalues)
        self.coef = df.reset_index()

    def summary(
        self,
        main=True,
        fit=True,
        ci=False,
        ssq=False,
        rmse=False,
        vif=False,
        test=None,
        dec=3,
    ) -> None:
        """
        Summarize output from a linear regression model

        parameters
        ----------
        ssq: Boolean; if True, include sum of squares
        vif: Boolean; if True, include variance inflation factors
        """
        if main:
            print("Linear regression (OLS)")
            print("Data                 :", self.name)
            print("Response variable    :", self.rvar)
            print("Explanatory variables:", ", ".join(self.evar))
            print(f"Null hyp.: the effect of x on {self.rvar} is zero")
            print(f"Alt. hyp.: the effect of x on {self.rvar} is not zero")

            df = self.coef.copy()
            df["coefficient"] = df["coefficient"].round(dec)
            df["std.error"] = df["std.error"].round(dec)
            df["t.value"] = df["t.value"].round(dec)
            df["p.value"] = ifelse(
                df["p.value"] < 0.001, "< .001", df["p.value"].round(dec)
            )
            df["index"] = df["index"].str.replace("[T.", "[", regex=False)
            df = df.set_index("index")
            df.index.name = None
            print(f"\n{df.to_string()}")
            print("\nSignif. codes:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1")

        if fit:
            print(f"\n{model_fit(self.fitted, dec=dec)}")

        if ci:
            print("\nConfidence intervals:")
            df = coef_ci(self.fitted, dec=dec)
            print(f"\n{df.to_string()}")

        if ssq:
            print("\nSum of squares:")
            index = ["Regression", "Error", "Total"]
            sum_of_squares = [
                self.fitted.ess,
                self.fitted.ssr,
                self.fitted.centered_tss,
            ]
            sum_of_squares = pd.DataFrame(index=index).assign(
                df=format_nr(
                    [
                        self.fitted.df_model,
                        self.fitted.df_resid,
                        self.fitted.df_model + self.fitted.df_resid,
                    ],
                    dec=0,
                ),
                SS=format_nr(sum_of_squares, dec=0),
            )
            print(f"\n{sum_of_squares.to_string()}")

        if rmse:
            print("\nRoot Mean Square Error (RMSE):")
            rmse = (self.fitted.ssr / self.fitted.nobs) ** 0.5
            print(round(rmse, dec))

        if vif:
            if self.evar is None or len(self.evar) < 2:
                print("\nVariance Inflation Factors cannot be calculated")
            else:
                print("\nVariance inflation factors:")
                print(f"\n{calc_vif(self.fitted).to_string()}")

        if test is not None and len(test) > 0:
            self.f_test(test=test, dec=dec)

    def predict(
        self, data=None, cmd=None, data_cmd=None, ci=False, conf=0.95
    ) -> pd.DataFrame:
        """
        Predict values for a linear regression model
        """
        if data is None:
            data = self.data
        data = data.loc[:, self.evar].copy()
        if data_cmd is not None:
            for k, v in data_cmd.items():
                data[k] = v
        elif cmd is not None:
            cmd = {k: ifelse(isinstance(v, str), [v], v) for k, v in cmd.items()}
            data = sim_prediction(data=data, vary=cmd)

        if ci:
            if data_cmd is not None:
                raise ValueError(
                    "Confidence intervals not available when using the Data & Command option"
                )
            else:
                return pd.concat(
                    [data, predict_ci(self.fitted, data, conf=conf)], axis=1
                )
        else:
            return data.assign(prediction=self.fitted.predict(data))

    def plot(
        self,
        plots="dist",
        data=None,
        alpha=0.05,
        nobs: int = 1000,
        intercept=False,
        incl=None,
        excl=None,
        incl_int=[],
        fix=True,
        hline=False,
        figsize=None,
    ) -> None:
        """
        Plots for a linear regression model
        """
        data = self.data[[self.rvar] + self.evar].copy()
        if "dist" in plots:
            distr_plot(data)
        if "corr" in plots:
            cr = correlation(data)
            cr.plot(nobs=nobs, figsize=figsize)
        if "scatter" in plots:
            scatter_plot(self.fitted, data, nobs=nobs, figsize=figsize)
        if "dashboard" in plots:
            reg_dashboard(self.fitted, nobs=nobs)
        if "residual" in plots:
            residual_plot(self.fitted, data, nobs=nobs, figsize=figsize)
        if "pred" in plots:
            pred_plot_sm(
                self.fitted,
                data=ifelse(data is None, self.data, data),
                incl=incl,
                excl=ifelse(excl is None, [], excl),
                incl_int=incl_int,
                fix=fix,
                hline=hline,
                nnv=20,
                minq=0.025,
                maxq=0.975,
            )
        if "vimp" in plots:
            vimp_plot_sm(
                self.fitted,
                data=ifelse(data is None, self.data, data),
                rep=10,
                ax=None,
                ret=False,
            )
        if "coef" in plots:
            coef_plot(
                self.fitted,
                alpha=alpha,
                intercept=intercept,
                incl=incl,
                excl=excl,
                figsize=figsize,
            )

    def f_test(self, test=None, dec=3) -> None:
        """
        F-test for competing models

        Parameters
        ----------
        test : list
            List of strings; contains the names of the columns of data to be tested
        """
        evar = setdiff(self.evar, test)
        if self.ivar is not None and len(self.ivar) > 0:
            sint = setdiff(self.ivar, test)
            test += [s for t in test for s in sint if f"I({t}" not in s and t in s]
            sint = setdiff(sint, test)
        else:
            sint = []

        form = f"{self.rvar} ~ "
        if len(evar) == 0 and len(sint) == 0:
            form += "1"
        else:
            form += f"{' + '.join(evar + sint)}"
        pattern = r"(\[T\.[^\]]*\])\:"

        # ensure constraints are unique
        hypothesis = list(
            set(
                [
                    f"({c} = 0)"
                    for c in self.fitted.model.exog_names
                    for v in test
                    if f"{v}:" in c
                    or f":{v}" in c
                    or f"{v}[T." in c
                    or v == c
                    or v == re.sub(pattern, ":", c)
                ]
            )
        )

        print(f"\nModel 1: {form}")
        print(f"Model 2: {self.form}")
        out = self.fitted.f_test(hypothesis)

        r2_sub = self.fitted.rsquared - (
            len(hypothesis) * out.fvalue * (1 - self.fitted.rsquared)
        ) / (self.fitted.nobs - self.fitted.df_model - 1)

        pvalue = ifelse(out.pvalue < 0.001, "< .001", round(out.pvalue, dec))
        print(f"R-squared, Model 1 vs 2: {r2_sub:.3f} vs {self.fitted.rsquared:.3f}")
        print(
            f"F-statistic: {round(out.fvalue, dec)} df ({out.df_num:.0f}, {out.df_denom:.0f}), p.value {pvalue}"
        )



import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np
import pandas as pd
import polars as pl
from scipy import stats
from pyrsm.model import sig_stars
from pyrsm.utils import ifelse, check_dataframe
from typing import Union
from statsmodels.stats import multitest
import pyrsm.basics.utils as bu
import pyrsm.radiant.utils as ru


class compare_means:
    """
    Compare mean between numeric variables in a Pandas or Polars dataframe. See
    the notebook linked below for a worked example, including the web UI:

    https://github.com/vnijs/pyrsm/blob/main/examples/basics-compare-means.ipynb

    Parameters
    ----------
    data : Pandas or Polars dataframe with categorical variables or a
        dictionary with a single dataframe as the value and the
        name of the dataframe as the key
    var1: String; Name of the first numeric or categorical variable
    var2: String or list of Strings; Names of the numberic variables
    comb: List of string tuples; Comparisons to make. Will default to
        all possible combinations
    alt_hyp: String; Alternative hypothesis to use "two-sided", "less",
        or "greater"
    conf: float; Confidence level (e.g., .95 or .9)
    sample_type: String; Are samples independent ("independent") or not ("paired")
    adjust: String; Adjustment for multiple testing (None or "bonferoni" or
        other options provided but statsmodels.stats.multitest.multipletests)
    test_type: String; "t-test" or "wilcox"

    Examples
    --------

    import pandas as pd
    import pyrsm as rsm
    salary, salary_description = rsm.load_data(pkg="basics", name="salary")
    cm = rsm.basics.compare_means({"salary": salary}, var1="rank", var2="salary", alt_hyp="less")
    cm.summary()
    """

    def __init__(
        self,
        data: pd.DataFrame | pl.DataFrame | dict[str, pd.DataFrame | pl.DataFrame],
        var1: str,
        var2: str,
        comb: list[tuple[str, str]] = [],
        alt_hyp: str = "two-sided",
        conf: float = 0.95,
        sample_type: str = "independent",
        adjust: str = None,
        test_type: str = "t-test",
    ):
        if isinstance(data, dict):
            self.name = list(data.keys())[0]
            self.data = data[self.name]
        else:
            self.data = data
            self.name = "Not provided"

        self.data = check_dataframe(self.data)
        self.var1 = var1
        self.var2 = var2
        self.comb = comb

        var1_series = self.data[self.var1]
        if pd.api.types.is_numeric_dtype(var1_series):
            var2 = ifelse(
                isinstance(var2, str),
                [var2],
                ifelse(isinstance(var2, tuple), list(var2), var2),
            )
            if (
                len(self.var2) > 1
                or len([v for v in self.comb if self.var1 in v]) > 0
                or var1_series.nunique() > 10
            ):
                self.data = self.data.loc[:, [self.var1] + var2].melt()
                self.var1 = "variable"
                self.var2 = "value"

        self.data[self.var1] = self.data[self.var1].astype("category")
        self.levels = list(self.data[self.var1].cat.categories)

        if len(self.comb) == 0:
            self.comb = ru.iterms(self.levels)

        self.alt_hyp = alt_hyp
        self.conf = conf
        self.alpha = 1 - self.conf
        self.sample_type = sample_type
        self.adjust = adjust
        self.test_type = test_type

        def welch_dof(v1: str, v2: str) -> float:
            # stats.ttest_ind uses Welch's t-test when equal_var=False
            # but does not return the degrees of freedom
            x = self.data.loc[self.data[self.var1] == v1, self.var2]
            y = self.data.loc[self.data[self.var1] == v2, self.var2]
            if x.size == 0 or y.size == 0:  # address division by zero
                return np.nan
            dof = (x.var() / x.size + y.var() / y.size) ** 2 / (
                (x.var() / x.size) ** 2 / (x.size - 1)
                + (y.var() / y.size) ** 2 / (y.size - 1)
            )

            return dof

        descriptive_stats = []
        for lev in self.levels:
            subset = self.data.loc[self.data[self.var1] == lev, self.var2]
            mean = np.nanmean(subset)
            n_missing = subset.isna().sum()
            n = len(subset) - n_missing
            sd = subset.std()
            se = subset.sem()
            tscore = stats.t.ppf((1 + self.conf) / 2, n - 1)
            me = (tscore * se).real
            row = [lev, mean, n, n_missing, sd, se, me]
            descriptive_stats.append(row)

        self.descriptive_stats = pd.DataFrame(
            descriptive_stats,
            columns=[self.var1, "mean", "n", "n_missing", "sd", "se", "me"],
        )

        if self.alt_hyp == "less":
            alt_hyp_sign = "less than"
        elif self.alt_hyp == "two-sided":
            alt_hyp_sign = "not equal to"
        else:
            alt_hyp_sign = "greater than"

        comp_stats = []
        for c in self.comb:
            v1, v2 = c.split(":")
            null_hyp = f"{v1} = {v2}"
            alt_hyp = f"{v1} {alt_hyp_sign} {v2}"
            diff = np.nanmean(
                self.data[self.data[self.var1] == v1][self.var2]
            ) - np.nanmean(self.data[self.data[self.var1] == v2][self.var2])

            x = self.data.loc[self.data[self.var1] == v1, self.var2]
            y = self.data.loc[self.data[self.var1] == v2, self.var2]
            if x.size != y.size and self.sample_type == "paired":
                raise ValueError(
                    """The two samples must have the same size for a paired
                    sample test. Choose independent samples instead."""
                )

            if self.test_type == "t-test":
                if self.sample_type == "independent":
                    result = stats.ttest_ind(
                        x,
                        y,
                        equal_var=False,
                        nan_policy="omit",
                        alternative=self.alt_hyp,
                    )
                else:
                    result = stats.ttest_rel(
                        x, y, nan_policy="omit", alternative=self.alt_hyp
                    )
            elif self.test_type == "wilcox":
                if self.sample_type == "independent":
                    result = stats.ranksums(x, y, alternative=self.alt_hyp)
                else:
                    result = stats.wilcoxon(
                        x, y, correction=True, alternative=self.alt_hyp
                    )

            t_val, p_val = result.statistic, result.pvalue
            se = diff / t_val
            df = welch_dof(v1, v2)

            if self.alt_hyp == "two-sided":
                tscore = stats.t.ppf((1 + self.conf) / 2, df)
            else:
                tscore = stats.t.ppf(self.conf, df)
            me = (tscore * se).real

            if self.alt_hyp == "less":
                ci = [-np.inf, diff + me]
            elif self.alt_hyp == "two-sided":
                ci = [diff - me, diff + me]
            else:
                ci = [diff - me, np.inf]

            comp_stats.append(
                [
                    null_hyp,
                    alt_hyp,
                    diff,
                    p_val,
                    se,
                    t_val,
                    df,
                    ci[0],
                    ci[1],
                    sig_stars([p_val])[0],
                ]
            )

        cl = bu.ci_label(self.alt_hyp, self.conf)
        self.comp_stats = pd.DataFrame(
            comp_stats,
            columns=[
                "Null hyp.",
                "Alt. hyp.",
                "diff",
                "p.value",
                "se",
                "t.value",
                "df",
                cl[0],
                cl[1],
                "",
            ],
        )

        if self.adjust is not None:
            if self.alt_hyp == "two-sided":
                alpha = self.alpha
            else:
                alpha = self.alpha * 2
            self.comp_stats["p.value"] = multitest.multipletests(
                self.comp_stats["p.value"], method=self.adjust, alpha=alpha
            )[1]
            self.comp_stats[""] = sig_stars(self.comp_stats["p.value"])

    def summary(self, extra=False, dec=3) -> None:
        print(f"Pairwise mean comparisons ({self.test_type})")
        print(f"Data      : {self.name}")
        print(f"Variables : {self.var1}, {self.var2}")
        print(f"Samples   : {self.sample_type}")
        print(f"Confidence: {self.conf}")
        print(f"Adjustment: {self.adjust}")

        print(self.descriptive_stats.round(dec).to_string(index=False))

        comp_stats = self.comp_stats.copy()
        if not extra:
            comp_stats = comp_stats.iloc[:, [0, 1, 2, 3, -1]]

        comp_stats["p.value"] = ifelse(
            comp_stats["p.value"] < 0.001, "< .001", round(comp_stats["p.value"], dec)
        )
        print(comp_stats.round(dec).to_string(index=False))
        print("\nSignif. codes:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1")

    def plot(self, plots: str = "scatter", nobs: int = None) -> None:
        if plots == "scatter":
            if (
                nobs is not None
                and nobs < self.data.shape[0]
                and nobs != np.Inf
                and nobs != -1
            ):
                data = self.data.copy().sample(nobs)
            else:
                data = self.data.copy()

            sns.swarmplot(data=data, x=self.var1, y=self.var2, alpha=0.5)

            # Get the unique categories and their indices
            categories = data[self.var1].cat.categories
            category_indices = {category: i for i, category in enumerate(categories)}

            category_means = data.groupby(self.var1, observed=False)[self.var2].mean()

            # Add a horizontal line for each category at the mean of the value for that category
            for category, mean in category_means.items():
                plt.hlines(
                    y=mean,
                    xmin=category_indices[category] - 0.3,
                    xmax=category_indices[category] + 0.3,
                    color="blue",
                    linestyle="--",
                    linewidth=2,
                    zorder=2,
                )

        elif plots == "box":
            sns.boxplot(data=self.data, x=self.var1, y=self.var2)
        elif plots == "density":
            sns.kdeplot(data=self.data, x=self.var2, hue=self.var1)
        elif plots == "bar":
            sns.barplot(
                data=self.data,
                y=self.var2,
                x=self.var1,
                # yerr=self.descriptive_stats["se"], # shapes don't align for some reason
            )
        else:
            print("Invalid plot type")



from starlette.applications import Starlette
from starlette.routing import Mount
from starlette.staticfiles import StaticFiles
from starlette.requests import Request as StarletteRequest
from pathlib import Path
from shiny import App, render, ui, reactive, Inputs, Outputs, Session, req
import webbrowser
import nest_asyncio
import uvicorn
import signal
import os
import io
import sys
import tempfile
import matplotlib.pyplot as plt
from contextlib import redirect_stdout, redirect_stderr
import pyrsm as rsm
import pyrsm.radiant.utils as ru
import pyrsm.basics.probability_calculator as pc

pc_type = {"values": "Values", "probs": "Probabilities"}
pc_dist = {v: k for k, v in pc.pc_dist.items()}


def ui_type(self):
    return ui.panel_well(
        ui.input_radio_buttons(
            id="type",
            label="Input type:",
            selected=self.state.get("type", "values"),
            choices=pc_type,
            inline=True,
        ),
        ui.panel_conditional(
            "input.type == 'values'",
            ru.make_side_by_side(
                ui.input_numeric(
                    "lb", "Lower bound:", value=self.state.get("lb", None)
                ),
                ui.input_numeric(
                    "ub", "Upper bound:", value=self.state.get("ub", None)
                ),
            ),
        ),
        ui.panel_conditional(
            "input.type == 'probs'",
            ru.make_side_by_side(
                ui.input_numeric(
                    "plb",
                    "Lower bound:",
                    value=self.state.get("plb", None),
                    min=0,
                    max=1,
                    step=0.05,
                ),
                ui.input_numeric(
                    "pub",
                    "Upper bound:",
                    value=self.state.get("pub", None),
                    min=0,
                    max=1,
                    step=0.05,
                ),
            ),
        ),
        ui.input_numeric("dec", "Decimals:", value=self.state.get("dec", 3)),
    )


def ui_summary(self):
    return ui.panel_well(
        ui.input_select(
            id="dist",
            label="Distribution:",
            selected=self.state.get("dist", "binom"),
            choices=pc_dist,
        ),
        ui.panel_conditional("input.dist == 'binom'", ui.output_ui("ui_pc_binom")),
        ui.panel_conditional("input.dist == 'chisq'", ui.output_ui("ui_pc_chisq")),
        ui.panel_conditional("input.dist == 'disc'", ui.output_ui("ui_pc_disc")),
        ui.panel_conditional("input.dist == 'expo'", ui.output_ui("ui_pc_expo")),
        ui.panel_conditional("input.dist == 'fdist'", ui.output_ui("ui_pc_fdist")),
        ui.panel_conditional("input.dist == 'lnorm'", ui.output_ui("ui_pc_lnorm")),
        ui.panel_conditional("input.dist == 'norm'", ui.output_ui("ui_pc_norm")),
        ui.panel_conditional("input.dist == 'pois'", ui.output_ui("ui_pc_pois")),
        ui.panel_conditional("input.dist == 'tdist'", ui.output_ui("ui_pc_tdist")),
        ui.panel_conditional("input.dist == 'unif'", ui.output_ui("ui_pc_unif")),
    )


# def ui_main_pc(height="500px", width="700px"):
def ui_main_pc():
    return ui.panel_well(
        ui.output_ui("show_summary_code"),
        ui.output_text_verbatim("summary"),
        ui.output_ui("plot_container"),
        id="pc_well",
    )


class basics_probability_calculator:
    def __init__(self, state=None, code=True, navbar=None) -> None:
        self.code = code
        if state is None:
            self.state = {}
        else:
            self.state = state
        if navbar is None:
            self.navbar = ()
        else:
            self.navbar = navbar

    def shiny_ui(self, request: StarletteRequest):
        return ui.page_navbar(
            ru.head_content(),
            ui.nav(
                "Basics > Probability calculator",
                ui.row(
                    ui.column(
                        3,
                        ui_summary(self),
                        ui_type(self),
                    ),
                    ui.column(8, ui_main_pc()),
                ),
            ),
            self.navbar,
            ru.ui_help(
                "https://github.com/vnijs/pyrsm/blob/main/examples/basics/basics-probability-calculator.ipynb",
                "Probability calculator example notebook",
            ),
            ru.ui_stop(),
            title="Radiant for Python",
            inverse=False,
            id="navbar_id",
        )

    def shiny_server(self, input: Inputs, output: Outputs, session: Session):
        def update_state():
            with reactive.isolate():
                ru.dct_update(self, input)

        session.on_ended(update_state)

        @output(id="ui_pc_binom")
        @render.ui
        def ui_pc_binom():
            return ru.make_side_by_side(
                ui.input_numeric(
                    "binom_n", label="n:", value=self.state.get("binom_n", 10), min=0
                ),
                ui.input_numeric(
                    "binom_p",
                    label="p:",
                    value=self.state.get("binom_p", 0.2),
                    min=0,
                    max=1,
                    step=0.01,
                ),
            )

        @output(id="ui_pc_chisq")
        @render.ui
        def ui_pc_chisq():
            return ui.input_numeric(
                "chisq_df",
                label="Degrees of freedom:",
                value=self.state.get("chisq_df", 1),
                min=1,
            )

        @output(id="ui_pc_disc")
        @render.ui
        def ui_pc_disc():
            return ui.panel_well(
                ru.input_return_text_area(
                    "disc_values",
                    label="Values:",
                    placeholder="Insert list [1, 3, 5]",
                    value=self.state.get("disc_values", "[1, 3, 5]"),
                ),
                ru.input_return_text_area(
                    "disc_probs",
                    label="Probabilities:",
                    placeholder="Insert list [1/4, 1/8, 5/8]",
                    value=self.state.get("disc_probs", "[1 / 4, 1 / 8, 5 / 8]"),
                ),
            )

        @output(id="ui_pc_expo")
        @render.ui
        def ui_pc_expo():
            return ui.input_numeric(
                "expo_rate", label="Rate:", value=self.state.get("expo_rate", 1), min=1
            )

        @output(id="ui_pc_fdist")
        @render.ui
        def ui_pc_fdist():
            return ru.make_side_by_side(
                ui.input_numeric(
                    "fdist_df1",
                    label="Degrees of freedom 1:",
                    value=self.state.get("fdist_df1", 10),
                    min=0,
                ),
                ui.input_numeric(
                    "fdist_df2",
                    label="Degrees of freedom 2:",
                    value=self.state.get("fdist_df2", 10),
                    min=0,
                ),
            )

        @output(id="ui_pc_lnorm")
        @render.ui
        def ui_pc_lnorm():
            return ru.make_side_by_side(
                ui.input_numeric(
                    "lnorm_mean",
                    label="Mean log:",
                    value=self.state.get("lnorm_mean", 0),
                ),
                ui.input_numeric(
                    "lnorm_stdev",
                    label="St. dev. log:",
                    value=self.state.get("lnorm_stdev", 1),
                    min=0,
                ),
            )

        @output(id="ui_pc_norm")
        @render.ui
        def ui_pc_norm():
            return ru.make_side_by_side(
                ui.input_numeric(
                    "norm_mean", label="Mean:", value=self.state.get("norm_mean", 0)
                ),
                ui.input_numeric(
                    "norm_stdev",
                    label="St. dev.:",
                    value=self.state.get("norm_stdev", 1),
                    min=0,
                ),
            )

        @output(id="ui_pc_pois")
        @render.ui
        def ui_pc_pois():
            return ui.input_numeric(
                "pois_lambda",
                label="Lambda:",
                value=self.state.get("pois_lambda", 0.5),
                min=0,
                step=0.1,
            )

        @output(id="ui_pc_tdist")
        @render.ui
        def ui_pc_tdist():
            return ui.input_numeric(
                "tdist_df",
                label="Degrees of freedom:",
                value=self.state.get("tdist_df", 10),
                min=1,
            )

        @output(id="ui_pc_unif")
        @render.ui
        def ui_pc_unif():
            return ru.make_side_by_side(
                ui.input_numeric(
                    "unif_min", label="Minimum:", value=self.state.get("unif_min", 0)
                ),
                ui.input_numeric(
                    "unif_max", label="Maximum:", value=self.state.get("unif_max", 1)
                ),
            )

        def estimation_code():
            dist = input.dist()
            ignore = [""]
            if dist == "binom":
                args = {
                    "n": input.binom_n(),
                    "p": input.binom_p(),
                }
            elif dist == "chisq":
                args = {"df": input.chisq_df()}
            elif dist == "disc":
                args = {
                    "v": input.disc_values(),
                    "p": input.disc_probs(),
                }
                ignore = ["v", "p"]
            elif dist == "expo":
                args = {"rate": input.expo_rate()}
            elif dist == "fdist":
                args = {
                    "df1": input.fdist_df1(),
                    "df2": input.fdist_df2(),
                }
            elif dist == "lnorm":
                args = {
                    "meanlog": input.lnorm_mean(),
                    "sdlog": input.lnorm_stdev(),
                }
            elif dist == "norm":
                args = {
                    "mean": input.norm_mean(),
                    "stdev": input.norm_stdev(),
                }
            elif dist == "pois":
                args = {"lamb": input.pois_lambda()}
            elif dist == "tdist":
                args = {"df": input.tdist_df()}
            elif dist == "unif":
                args = {"min": input.unif_min(), "max": input.unif_max()}
            else:
                pass

            # inputs in args must be available for calculation to work
            [req(v is not None) for v in args.values()]

            if input.type() == "values":
                args.update({"lb": input.lb(), "ub": input.ub()})
            else:
                if input.plb() is not None and (input.plb() < 0 or input.plb() > 1):
                    raise ValueError("All probabilities must be between 0 and 1")
                elif input.pub() is not None and (input.pub() < 0 or input.pub() > 1):
                    raise ValueError("All probabilities must be between 0 and 1")

                args.update({"plb": input.plb(), "pub": input.pub()})

            fun = getattr(
                rsm.basics.probability_calculator_functions, f"prob_{input.dist()}"
            )
            args_string = ru.drop_default_args(args, fun, ignore=ignore)
            return f"""rsm.basics.prob_calc("{dist}", {args_string})"""

        @reactive.Calc
        def estimate():
            return eval(estimation_code())

        def summary_code():
            args = {"dec": input.dec()}
            fun = getattr(
                rsm.basics.probability_calculator_functions,
                f"summary_prob_{input.dist()}",
            )
            args_string = ru.drop_default_args(args, fun)
            return f"""pc.summary({args_string})"""

        @output(id="show_summary_code")
        @render.text
        def show_summary_code():
            cmd = f"""import pyrsm as rsm\npc = {estimation_code()}\n{summary_code()}\npc.plot()"""
            return ru.code_formatter(cmd, self, input, session)

        @output(id="summary")
        @render.text
        def summary():
            out = io.StringIO()
            with redirect_stdout(out), redirect_stderr(out):
                pc = estimate()  # get estimation object into local scope
                eval(summary_code())
            return out.getvalue()

        @reactive.Calc
        def gen_plot():
            locals()["pc"] = estimate()
            eval("""pc.plot()""")
            fig = plt.gcf()
            width, height = fig.get_size_inches()  # Get the size in inches
            return fig, width * 96, height * 96

        @output(id="plot")
        @render.plot
        def plot():
            return gen_plot()[0]

        @output(id="plot_container")
        @render.ui
        def plot_container():
            req(estimate())
            width, height = gen_plot()[1:]
            return ui.output_plot("plot", height=f"{height}px", width=f"{width}px")

        # --- section standard for all apps ---
        # stops returning code if moved to utils
        @reactive.Effect
        @reactive.event(input.stop, ignore_none=True)
        async def stop_app():
            rsm.md(f"```python\n{self.stop_code}\n```")
            await session.app.stop()
            os.kill(os.getpid(), signal.SIGTERM)

            return self.stop_code


def prob_calc(
    code: bool = True,
    state: dict = None,
    host: str = "",
    port: int = 8000,
    log_level: str = "critical",
    debug: bool = False,
):
    """
    Launch a Radiant-for-Python app for compare means hypothesis testing
    """
    host = ru.set_host(host)
    rc = basics_probability_calculator(state=state, code=code)
    nest_asyncio.apply()
    webbrowser.open(f"http://{host}:{port}")
    print(f"Listening on http://{host}:{port}")
    ru.message()

    # redirect stdout and stderr to the temporary file
    if not debug:
        org_stdout = sys.stdout
        org_stderr = sys.stderr
        temp = tempfile.NamedTemporaryFile()
        temp_file = open(temp.name, "w")
        sys.stdout = temp_file
        sys.stderr = temp_file

    app = App(rc.shiny_ui, rc.shiny_server)
    www_dir = Path(__file__).parent.parent.parent / "radiant" / "www"
    app_static = StaticFiles(directory=www_dir, html=False)

    routes = [
        Mount("/www", app=app_static),
        Mount("/", app=app),
    ]

    uvicorn.run(
        Starlette(debug=debug, routes=routes),
        host=host,
        port=port,
        log_level=log_level,
    )

    if not debug:
        sys.stdout = org_stdout
        sys.stderr = org_stderr
        temp_file.close()


if __name__ == "__main__":
    prob_calc(debug=False)



from typing import Optional, Literal
import pandas as pd
import polars as pl
import matplotlib.pyplot as plt

# from sklearn.model_selection import GridSearchCV
# from sklearn.model_selection import StratifiedKFold
from sklearn.neural_network import MLPClassifier, MLPRegressor
from sklearn.inspection import PartialDependenceDisplay as pdp
from pyrsm.utils import ifelse, check_dataframe, setdiff
from pyrsm.model.model import (
    sim_prediction,
    convert_binary,
    evalreg,
    convert_to_list,
    reg_dashboard,
)
from pyrsm.model.perf import auc
from pyrsm.stats import scale_df
from .visualize import pred_plot_sk, vimp_plot_sk


class mlp:
    """
    Initialize Multi-layer Perceptron (NN) model

    Parameters
    ----------
    data: pandas DataFrame; dataset
    lev: String; name of the level in the response variable
    rvar: String; name of the column to be used as the response variable
    evar: List of strings; contains the names of the column of data to be used as the explanatory (target) variable
    hidden_layer_sizes: The number of neurons in the hidden layer and the number of hidden layers (e..g, (5,) for 5 neurons in 1 hidden layer, (5, 5) for 5 neurons in 2 hidden layers, etc.)
    activation: Activation function apply to transform for the nodes in the hidden layer
    solver: The solver used for weight optimization
    alpha: L2 penalty (regularization term) parameter

    **kwargs : Named arguments to be passed to the sklearn's Multi-layer Perceptron functions
    """

    def __init__(
        self,
        data: pd.DataFrame | pl.DataFrame | dict[str, pd.DataFrame | pl.DataFrame],
        rvar: Optional[str] = None,
        lev: Optional[str] = None,
        evar: Optional[list[str]] = None,
        hidden_layer_sizes: tuple = (5,),
        activation: Literal["identity", "logistic", "tanh", "relu"] = "tanh",
        solver: Literal["lbfgs", "sgd", "adam"] = "lbfgs",
        alpha: float = 0.0001,
        batch_size: float | str = "auto",
        learning_rate_init: float = 0.001,
        max_iter: int = 10_000,
        random_state: int = 1234,
        mod_type: Literal["regression", "classification"] = "classification",
        **kwargs,
    ) -> None:
        if isinstance(data, dict):
            self.name = list(data.keys())[0]
            self.data = data[self.name]
        else:
            self.data = data
            self.name = "Not provided"
        self.data = check_dataframe(self.data)
        self.rvar = rvar
        self.lev = lev
        self.evar = convert_to_list(evar)
        self.mod_type = mod_type
        self.hidden_layer_sizes = hidden_layer_sizes
        self.activation = activation
        self.solver = solver
        self.alpha = alpha
        self.batch_size = batch_size
        self.learning_rate_init = learning_rate_init
        self.max_iter = max_iter
        self.random_state = random_state
        self.kwargs = kwargs

        if self.mod_type == "classification":
            if self.lev is not None and self.rvar is not None:
                self.data[self.rvar] = convert_binary(self.data[self.rvar], self.lev)

            self.mlp = MLPClassifier(
                hidden_layer_sizes=self.hidden_layer_sizes,
                activation=self.activation,
                solver=self.solver,
                alpha=self.alpha,
                batch_size=self.batch_size,
                learning_rate_init=self.learning_rate_init,
                max_iter=self.max_iter,
                random_state=self.random_state,
                **kwargs,
            )
        else:
            self.mlp = MLPRegressor(
                hidden_layer_sizes=self.hidden_layer_sizes,
                activation=self.activation,
                solver=self.solver,
                alpha=self.alpha,
                batch_size=self.batch_size,
                learning_rate_init=self.learning_rate_init,
                max_iter=self.max_iter,
                random_state=self.random_state,
                **kwargs,
            )
        self.data_std, self.means, self.stds = scale_df(
            self.data[[rvar] + self.evar], sf=1, stats=True
        )
        # use drop_first=True
        self.data_onehot = pd.get_dummies(self.data_std[self.evar], drop_first=True)
        self.n_features = [len(evar), self.data_onehot.shape[1]]
        self.fitted = self.mlp.fit(self.data_onehot, self.data_std[self.rvar])
        self.nobs = self.data.dropna().shape[0]

    def summary(self, dec=3) -> None:
        """
        Summarize output from a Multi-layer Perceptron (NN) model
        """
        print("Multi-layer Perceptron (NN)")
        print(f"Data                 : {self.name}")
        print(f"Response variable    : {self.rvar}")
        if self.mod_type == "classification":
            print(f"Level                : {self.lev}")
        print(f"Explanatory variables: {', '.join(self.evar)}")
        print(
            f"Model type           : {ifelse(self.mod_type == 'classification', 'classification', 'regression')}"
        )
        print(f"Nr. of features      : ({self.n_features[0]}, {self.n_features[1]})")
        print(f"Nr. of observations  : {format(self.nobs, ',.0f')}")
        print(f"Hidden_layer_sizes   : {self.hidden_layer_sizes}")
        print(f"Activation function  : {self.activation}")
        print(f"Solver               : {self.solver}")
        print(f"Alpha                : {self.alpha}")
        print(f"Batch size           : {self.batch_size}")
        print(f"Learning rate        : {self.learning_rate_init}")
        print(f"Maximum iterations   : {self.max_iter}")
        print(f"random_state         : {self.random_state}")
        if self.mod_type == "classification":
            print(
                f"AUC                  : {round(auc(self.data[self.rvar], self.fitted.predict_proba(self.data_onehot)[:, -1]), dec)}"
            )
        else:
            print("Model fit            :")
            print(
                evalreg(
                    pd.DataFrame().assign(
                        rvar=self.data_std[[self.rvar]],
                        prediction=self.fitted.predict(self.data_onehot),
                    ),
                    "rvar",
                    "prediction",
                    dec=dec,
                )
                .T[2:]
                .rename(columns={0: " "})
                .T.to_string()
            )

        if len(self.kwargs) > 0:
            kwargs_list = [f"{k}={v}" for k, v in self.kwargs.items()]
            print(f"Extra arguments      : {', '.join(kwargs_list)}")

        print("\nRaw data             :")
        print(self.data[self.evar].head().to_string(index=False))

        print("\nEstimation data      :")
        print(self.data_onehot.head().to_string(index=False))

    def predict(
        self, data=None, cmd=None, data_cmd=None, scale=True, means=None, stds=None
    ) -> pd.DataFrame:
        """
        Predict probabilities or values for an MLP
        """
        if data is None:
            data = self.data.loc[:, self.evar].copy()
        else:
            data = data.loc[:, self.evar].copy()

        if data_cmd is not None and data_cmd != "":
            for k, v in data_cmd.items():
                data[k] = v
        elif cmd is not None and cmd != "":
            cmd = {k: ifelse(isinstance(v, str), [v], v) for k, v in cmd.items()}
            data = sim_prediction(data=data, vary=cmd)

        if scale or (means is not None and stds is not None):
            if means is not None and stds is not None:
                data_std = scale_df(data, sf=1, means=means, stds=stds)
            else:
                # scaling the full dataset by the means used during estimation
                data_std = scale_df(data, sf=1, means=self.means, stds=self.stds)

            data_onehot = pd.get_dummies(data_std, drop_first=False)
        else:
            data_onehot = pd.get_dummies(data, drop_first=False)
        if data_onehot.shape[1] != self.data_onehot.shape[1]:
            data_onehot_missing = pd.DataFrame(
                {
                    k: [False] * data_onehot.shape[0]
                    for k in setdiff(self.data_onehot.columns, data_onehot.columns)
                }
            )
            data_onehot = pd.concat([data_onehot, data_onehot_missing], axis=1)
            data_onehot = data_onehot[self.data_onehot.columns]

        if self.mod_type == "classification":
            data["prediction"] = self.fitted.predict_proba(data_onehot)[:, -1]
        else:
            data["prediction"] = (
                self.fitted.predict(data_onehot) * self.stds[self.rvar]
                + self.means[self.rvar]
            )

        return data

    def plot(
        self,
        plots: Literal["pred", "pdp", "vimp", "dashboard"] = "pred",
        incl=None,
        excl=None,
        incl_int=[],
        nobs: int = 1000,
        fix=True,
        hline=False,
        figsize=None,
    ) -> None:
        """
        Plots for a Multi-layer Perceptron model (NN)
        """
        if "pred" in plots:
            pred_plot_sk(
                self.fitted,
                # self.data[[self.rvar] + self.evar],
                self.data_std[[self.rvar] + self.evar],
                self.rvar,
                incl=incl,
                excl=ifelse(excl is None, [], excl),
                incl_int=incl_int,
                fix=fix,
                hline=hline,
                nnv=40,
                minq=0.025,
                maxq=0.975,
            )

        if "pdp" in plots:
            if figsize is None:
                figsize = (8, len(self.data_onehot.columns) * 2)
            fig, ax = plt.subplots(figsize=figsize)
            ax.set_title("Partial Dependence Plots")
            fig = pdp.from_estimator(
                self.fitted, self.data_onehot, self.data_onehot.columns, ax=ax, n_cols=2
            )

        if "vimp" in plots:
            vimp_plot_sk(
                self.fitted,
                self.data_onehot,
                self.data[self.rvar],
                rep=5,
                ax=None,
                ret=False,
            )

        if "dashboard" in plots and self.mod_type == "regression":
            model = self.fitted
            model.fittedvalues = self.predict()["prediction"]
            model.resid = self.data[self.rvar] - model.fittedvalues
            model.model = pd.DataFrame({"endog": self.data[self.rvar]})
            reg_dashboard(model, nobs=nobs)



from pathlib import Path
from shiny import App
from starlette.applications import Starlette
from starlette.routing import Mount
from starlette.staticfiles import StaticFiles
from pyrsm.radiant import utils as ru
from pyrsm.radiant.data.data_view import data_view
from pyrsm.radiant.basics.probability_calculator import basics_probability_calculator
from pyrsm.radiant.basics.single_mean import basics_single_mean
from pyrsm.radiant.basics.single_prop import basics_single_prop
from pyrsm.radiant.basics.compare_means import basics_compare_means
from pyrsm.radiant.basics.compare_props import basics_compare_props
from pyrsm.radiant.basics.cross_tabs import basics_cross_tabs
from pyrsm.radiant.basics.goodness import basics_goodness
from pyrsm.radiant.basics.correlation import basics_correlation
from pyrsm.radiant.model.regress import model_regress
from pyrsm.radiant.model.logistic import model_logistic
import nest_asyncio
import uvicorn
import webbrowser
import sys
import tempfile


def radiant(
    data_dct: dict = None,
    descriptions_dct: dict = None,
    state: dict = None,
    code: bool = True,
    # host: str = "0.0.0.0",
    host: str = "localhost",
    port: int = 8000,
    log_level: str = "warning",
    debug: bool = False,
):
    """
    Launch the Radiant-for-Python app
    """

    # redirect stdout and stderr to the temporary file
    if not debug:
        org_stdout = sys.stdout
        org_stderr = sys.stderr
        temp = tempfile.NamedTemporaryFile()
        temp_file = open(temp.name, "w")
        sys.stdout = temp_file
        sys.stderr = temp_file

    data_dct, descriptions_dct = ru.get_dfs(pkg="model", name="diamonds")
    data_dct.update({"diamonds100": data_dct["diamonds"].sample(100)})

    # main path to www folder
    www_dir = Path(__file__).parent.parent / "radiant" / "www"

    # serve static files
    app_static = StaticFiles(directory=www_dir, html=False)

    # data > view
    data_dct, descriptions_dct = ru.get_dfs(pkg="data")
    rc = data_view(
        data_dct, descriptions_dct, state=None, code=True, navbar=ru.radiant_navbar()
    )
    app_data = App(rc.shiny_ui, rc.shiny_server, debug=False)

    # probability calculator mean app
    rc = basics_probability_calculator(
        state=None, code=True, navbar=ru.radiant_navbar()
    )
    app_pc = App(rc.shiny_ui, rc.shiny_server, debug=False)

    # single mean app
    data_dct, descriptions_dct = ru.get_dfs(pkg="basics", name="demand_uk")
    rc = basics_single_mean(
        data_dct, descriptions_dct, state=None, code=True, navbar=ru.radiant_navbar()
    )
    app_sm = App(rc.shiny_ui, rc.shiny_server, debug=False)

    # compare means app
    data_dct, descriptions_dct = ru.get_dfs(pkg="basics", name="salary")
    rc = basics_compare_means(
        data_dct, descriptions_dct, state=None, code=True, navbar=ru.radiant_navbar()
    )
    app_cm = App(rc.shiny_ui, rc.shiny_server, debug=False)

    # single prop
    data_dct, descriptions_dct = ru.get_dfs(pkg="basics", name="consider")
    rc = basics_single_prop(
        data_dct, descriptions_dct, state=None, code=True, navbar=ru.radiant_navbar()
    )
    app_sp = App(rc.shiny_ui, rc.shiny_server, debug=False)

    # compare props
    data_dct, descriptions_dct = ru.get_dfs(pkg="data", name="titanic")
    rc = basics_compare_props(
        data_dct, descriptions_dct, state=None, code=True, navbar=ru.radiant_navbar()
    )
    app_cp = App(rc.shiny_ui, rc.shiny_server, debug=False)

    # cross-tabs app
    data_dct, descriptions_dct = ru.get_dfs(pkg="basics", name="newspaper")
    rc = basics_cross_tabs(
        data_dct, descriptions_dct, state=None, code=True, navbar=ru.radiant_navbar()
    )
    app_ct = App(rc.shiny_ui, rc.shiny_server, debug=False)

    # goodness app
    data_dct, descriptions_dct = ru.get_dfs(pkg="basics", name="newspaper")
    rc = basics_goodness(
        data_dct, descriptions_dct, state=None, code=True, navbar=ru.radiant_navbar()
    )
    app_gf = App(rc.shiny_ui, rc.shiny_server, debug=False)

    # correlation app
    data_dct, descriptions_dct = ru.get_dfs(pkg="basics", name="salary")
    rc = basics_correlation(
        data_dct, descriptions_dct, state=None, code=True, navbar=ru.radiant_navbar()
    )
    app_cr = App(rc.shiny_ui, rc.shiny_server, debug=False)

    # linear regression app
    data_dct, descriptions_dct = ru.get_dfs(pkg="model", name="diamonds")
    data_dct.update({"diamonds100": data_dct["diamonds"].sample(100)})

    descriptions_dct.update({"diamonds100": descriptions_dct["diamonds"]})
    rc = model_regress(
        data_dct, descriptions_dct, state=None, code=True, navbar=ru.radiant_navbar()
    )
    app_regress = App(rc.shiny_ui, rc.shiny_server)

    # logistic regression app
    data_dct, descriptions_dct = ru.get_dfs(pkg="model", name="titanic")
    rc = model_logistic(
        data_dct, descriptions_dct, state=None, code=True, navbar=ru.radiant_navbar()
    )
    app_logistic = App(rc.shiny_ui, rc.shiny_server)

    # ---- combine apps ----
    routes = [
        Mount("/basics/prob-calc/", app=app_pc),
        Mount("/basics/single-mean/", app=app_sm),
        Mount("/basics/compare-means/", app=app_cm),
        Mount("/basics/single-prop/", app=app_sp),
        Mount("/basics/compare-props/", app=app_cp),
        Mount("/basics/cross-tabs/", app=app_ct),
        Mount("/basics/goodness/", app=app_gf),
        Mount("/basics/correlation/", app=app_cr),
        Mount("/models/regress/", app=app_regress),
        Mount("/models/logistic/", app=app_logistic),
        Mount("/www/", app=app_static),
        Mount("/", app=app_data),  # MUST BE LAST!!!!
    ]

    nest_asyncio.apply()
    webbrowser.open(f"http://{host}:{port}")
    print(f"Listening on http://{host}:{port}")
    ru.message()

    uvicorn.run(
        Starlette(debug=debug, routes=routes),
        host=host,
        port=port,
        log_level=log_level,
    )

    if not debug:
        sys.stdout = org_stdout
        sys.stderr = org_stderr
        temp_file.close()


if __name__ == "__main__":
    radiant(debug=False)



from typing import Optional, Literal
import pandas as pd
import polars as pl
import matplotlib.pyplot as plt
from math import sqrt, log2

# from sklearn.model_selection import GridSearchCV
# from sklearn.model_selection import StratifiedKFold
# from sklearn import metrics, tree
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.inspection import PartialDependenceDisplay as pdp
from pyrsm.utils import ifelse, check_dataframe, setdiff
from pyrsm.model.model import (
    sim_prediction,
    convert_binary,
    evalreg,
    convert_to_list,
    conditional_get_dummies,
    reg_dashboard,
)
from pyrsm.model.perf import auc
from .visualize import pred_plot_sk, vimp_plot_sk


class rforest:
    """
    Initialize random forest model

    Parameters
    ----------
    data: pandas DataFrame; dataset
    lev: String; name of the level in the response variable
    rvar: String; name of the column to be used as the response variable
    evar: List of strings; contains the names of the column of data to be used as the explanatory (target) variable
    n_estimators: The number of trees in the forest
    max_features: The number of features to consider when looking for the best split
    max_samples: The number of samples to draw from the data to train each tree
    oob_score: Whether to use out-of-bag samples to estimate the generalization accuracy
    random_state: Random seed used when bootstrapping data samples
    mod_type: String; type of model to be used (classification or regression)
    **kwargs : Named arguments to be passed to the sklearn's Random Forest functions
    """

    def __init__(
        self,
        data: pd.DataFrame | pl.DataFrame | dict[str, pd.DataFrame | pl.DataFrame],
        rvar: Optional[str] = None,
        lev: Optional[str] = None,
        evar: Optional[list[str]] = None,
        n_estimators: int = 100,
        min_samples_leaf: float | int = 1,
        max_features: float | int | Literal["sqrt", "log2"] = "sqrt",
        max_samples: float = 1.0,
        oob_score: bool = True,
        random_state: int = 1234,
        mod_type: Literal["regression", "classification"] = "classification",
        **kwargs,
    ) -> None:
        if isinstance(data, dict):
            self.name = list(data.keys())[0]
            self.data = data[self.name]
        else:
            self.data = data
            self.name = "Not provided"
        self.data = check_dataframe(self.data)
        self.rvar = rvar
        self.lev = lev
        self.evar = convert_to_list(evar)
        self.mod_type = mod_type
        self.oob_score = oob_score
        self.n_estimators = n_estimators
        self.min_samples_leaf = min_samples_leaf
        self.max_features = max_features
        self.max_samples = max_samples
        self.random_state = random_state
        self.kwargs = kwargs

        if self.mod_type == "classification":
            if self.lev is not None and self.rvar is not None:
                self.data[self.rvar] = convert_binary(self.data[self.rvar], self.lev)

            self.rf = RandomForestClassifier(
                n_estimators=self.n_estimators,
                min_samples_leaf=self.min_samples_leaf,
                max_features=self.max_features,
                oob_score=self.oob_score,
                random_state=self.random_state,
                **kwargs,
            )
        else:
            self.rf = RandomForestRegressor(
                n_estimators=self.n_estimators,
                min_samples_leaf=self.min_samples_leaf,
                max_features=self.max_features,
                oob_score=self.oob_score,
                random_state=self.random_state,
                **kwargs,
            )
        # only use drop_first=True for a decision tree where the categorical
        # variables are only binary
        # self.data_onehot = pd.get_dummies(self.data[evar], drop_first=drop_first)
        self.data_onehot = conditional_get_dummies(self.data[evar])
        self.n_features = [len(evar), self.data_onehot.shape[1]]
        self.fitted = self.rf.fit(self.data_onehot, self.data[self.rvar])
        self.nobs = self.data.dropna().shape[0]

    def summary(self, dec=3) -> None:
        """
        Summarize output from a Random Forest model
        """
        print("Random Forest")
        print(f"Data                 : {self.name}")
        print(f"Response variable    : {self.rvar}")
        if self.mod_type == "classification":
            print(f"Level                : {self.lev}")
        print(f"Explanatory variables: {', '.join(self.evar)}")
        print(f"OOB                  : {self.oob_score}")
        print(
            f"Model type           : {ifelse(self.mod_type == 'classification', 'classification', 'regression')}"
        )
        if self.max_features == "sqrt":
            # round down
            nr = int(sqrt(self.n_features[1]))
        elif self.max_features == "log2":
            # round down
            nr = int(log2(self.n_features[1]))
        elif self.max_features is None or self.max_features == "":
            nr = int(self.n_features[1])
        else:
            nr = int(self.max_features)
        print(f"Nr. of features      : ({self.n_features[0]}, {self.n_features[1]})")
        print(f"Nr. of observations  : {format(self.nobs, ',.0f')}")
        print(f"max_features         : {self.max_features} ({int(nr)})"),
        print(f"n_estimators         : {self.n_estimators}")
        print(f"min_samples_leaf     : {self.min_samples_leaf}")
        print(f"random_state         : {self.random_state}")
        if self.mod_type == "classification":
            cpred = self.fitted.oob_decision_function_[:, 1]
            print(
                f"AUC                  : {round(auc(self.data[self.rvar], cpred), dec)}"
            )
        else:
            print("Model fit            :")
            print(
                evalreg(
                    pd.DataFrame().assign(
                        rvar=self.data[[self.rvar]],
                        prediction=self.fitted.oob_prediction_,
                    ),
                    "rvar",
                    "prediction",
                    dec=dec,
                )
                .T[2:]
                .rename(columns={0: " "})
                .T.to_string()
            )

        if len(self.kwargs) > 0:
            kwargs_list = [f"{k}={v}" for k, v in self.kwargs.items()]
            print(f"Extra arguments      : {', '.join(kwargs_list)}")
        print("\nEstimation data      :")
        print(self.data_onehot.head().to_string(index=False))

    def predict(self, data=None, cmd=None, data_cmd=None) -> pd.DataFrame:
        """
        Predict probabilities or values for a random forest model
        """
        if (
            data is None
            and self.oob_score
            and (data_cmd is None or data_cmd == "")
            and (cmd is None or cmd == "")
        ):
            data = self.data.loc[:, self.evar].copy()
            if self.mod_type == "classification":
                pred = self.fitted.oob_decision_function_[:, 1]
            else:
                pred = self.fitted.oob_prediction_

            return data.assign(prediction=pred)
        elif data is None:
            data = self.data.loc[:, self.evar].copy()
        else:
            data = data.loc[:, self.evar].copy()

        if data_cmd is not None and data_cmd != "":
            for k, v in data_cmd.items():
                data[k] = v
        elif cmd is not None and cmd != "":
            cmd = {k: ifelse(isinstance(v, str), [v], v) for k, v in cmd.items()}
            data = sim_prediction(data=data, vary=cmd)

        # not dropping the first level of the categorical variables
        data_onehot = pd.get_dummies(data, drop_first=False)
        if data_onehot.shape[1] != self.data_onehot.shape[1]:
            data_onehot_missing = pd.DataFrame(
                {
                    k: [False] * data_onehot.shape[0]
                    for k in setdiff(self.data_onehot.columns, data_onehot.columns)
                }
            )
            data_onehot = pd.concat([data_onehot, data_onehot_missing], axis=1)
            data_onehot = data_onehot[self.data_onehot.columns]

        if self.mod_type == "classification":
            return data.assign(prediction=self.fitted.predict_proba(data_onehot)[:, -1])

        else:
            return data.assign(prediction=self.fitted.predict(data_onehot))

    def plot(
        self,
        plots: Literal["pred", "pdp", "vimp"] = "pred",
        incl=None,
        excl=None,
        incl_int=[],
        nobs: int = 1000,
        fix=True,
        hline=False,
        figsize=None,
    ) -> None:
        """
        Plots for a random forest model model
        """
        if "pred" in plots:
            pred_plot_sk(
                self.fitted,
                self.data[self.evar + [self.rvar]],
                self.rvar,
                incl=incl,
                excl=[],
                incl_int=incl_int,
                fix=fix,
                hline=hline,
                nnv=20,
                minq=0.025,
                maxq=0.975,
            )

        if "pdp" in plots:
            if figsize is None:
                figsize = (8, len(self.data_onehot.columns) * 2)
            fig, ax = plt.subplots(figsize=figsize)
            ax.set_title("Partial Dependence Plots")
            fig = pdp.from_estimator(
                self.fitted, self.data_onehot, self.data_onehot.columns, ax=ax, n_cols=2
            )

        if "vimp" in plots:
            vimp_plot_sk(
                self.fitted,
                self.data_onehot,
                self.data[self.rvar],
                rep=5,
                ax=None,
                ret=False,
            )

        if "dashboard" in plots and self.mod_type == "regression":
            model = self.fitted
            model.fittedvalues = self.predict()["prediction"]
            model.resid = self.data[self.rvar] - model.fittedvalues
            model.model = pd.DataFrame({"endog": self.data[self.rvar]})
            reg_dashboard(model, nobs=nobs)
